# imports
import numpy as np
import pandas as pd
import networkx as nx
from scipy.stats import binom
from random import randint, sample
from math import floor, exp, log, ceil, sqrt
from numpy.random import exponential as Exp
from copy import deepcopy
from .utilities import *

H1 = nx.Graph()
H0 = nx.Graph()
H = nx.Graph()
ldist = []
ldist1 = []
ldist0 = []
tdist = []
tdist1 = []
tdist0 = []
ddist1 = []
ddist0 = []
ddist = []
pairwise_segments = []

### random walks ###

def walk_variant_backward(s, p0, Ne, random_walk = False, one_step_model = 'a', tau0 = 0, sv=-0.01, ploidy = 2):
    '''Variant frequencies backward in time

    Parameters
    ----------
    s : float
        Selection coefficient
    p0 : float
        Variant frequency at generation 0
    Ne : dict
        Effective population sizes
    random_walk : bool
        True for random walk
    one_step_model : str
        'm', 'a', 'd', or 'r'
    tau0 : int
        Generation when neutrality begins
    sv: float
        Allele frequency of standing variation
        (Default -0.01 will assume de novo sweep)
    ploidy : int
        1 for haploid or 2 for diploid


    Returns
    -------
    tuple
        NumPy arrays for frequencies and sizes
    '''

    # local functions
    assert ploidy in [1,2]
    assert one_step_model in ['m','a','r','d']
    assert p0 <= 1
    assert p0 >= 0
    assert sv < 1
    def haploid_bwd(p, s): # haploid is same as multiplicative (Felsenstein, 2017)
        return p / (1 + s - s * p)
    def multiplicative_bwd(p, s):
        num = 1
        dnm = 1 + (1 - p) * s
        return p * num / dnm
    def dominant_bwd(p, s):
        if s <= 0:
            return p
        a = p * s
        b = 1 + s - 2 * p * s
        c = - p
        qf = - b + sqrt((b ** 2) - 4 * a * c)
        qf = qf / 2 / a
        return qf
    def additive_bwd(p, s):
        if s <= 0:
            return p
        a = s
        b = 1 + s - 2 * p * s
        c = - p
        qf = - b + sqrt((b ** 2) - 4 * a * c)
        qf = qf / 2 / a
        return qf
    def recessive_bwd(p, s):
        if s <= 0:
            return p
        a = (1 - p) * s
        b = 1
        c = - p
        qf = - b + sqrt((b ** 2) - 4 * a * c)
        qf = qf / 2 / a
        return qf

    # one step calculation
    if ploidy == 1:
        one_step = haploid_bwd
    else:
        if one_step_model == 'a':
            one_step = additive_bwd
        elif one_step_model == 'r':
            one_step = recessive_bwd
        elif one_step_model == 'd':
            one_step = dominant_bwd
        else:
            one_step = multiplicative_bwd

    # initialize
    ps = [] # frequencies
    xs = [] # variants
    Ns = [] # sizes
    t = floor(tau0)
    p = p0
    N = Ne[0]
    x = floor(p * ploidy * N)
    Ns.append(N)
    xs.append(x)
    ps.append(p)

    if random_walk: # random walk

        for G in range(1, max(Ne.keys())+1):
            try: # population size change
                N = Ne[G]
            except KeyError:
                pass
            if G > t:
                p = one_step(p, s)
            x = int(binom.rvs(int(ploidy * N), p))
            p = x / ploidy / N
            if x < 1:
                break
            if p >= 1:
                break
            if p <= sv:
                s = 0
            ps.append(p)
            xs.append(x)
            Ns.append(N)

        return np.array([ps, Ns, xs], dtype=float) # numpy-ify

    else: # deterministic

        for G in range(1, max(Ne.keys())+1):
            try: # population size change
                N = Ne[G]
            except KeyError:
                pass
            if G > t:
                p = one_step(p, s)
            x = floor(p * ploidy * N)
            if x < 1:
                break
            if p >= 1:
                break
            if p <= sv:
                s = 0
            Ns.append(N)
            xs.append(x)
            ps.append(p)

        return np.array([np.array(ps), np.array(Ns), np.array(xs)]) # numpy-ify

def walk_variant_forward(s, pG, Ne, random_walk = False, one_step_model = 'a', tau0 = 0, ploidy = 2):
    '''Variant frequencies forward in time

    Parameters
    ----------
    s : float
        Selection coefficient
    pG : float
        Variant frequency at maximum generation
    Ne : dict
        Effective population sizes
    random_walk : bool
        True for random walk
    one_step_model : str
        'm', 'a', 'd', or 'r'
    tau0 : int
        Generation when neutrality begins
    ploidy : int
        1 for haploid or 2 for diploid

    Returns
    -------
    tuple
        NumPy arrays for frequencies and sizes
    '''

    # local functions
    assert ploidy in [1,2]
    def haploid_fwd(p, s): # haploid is same as multiplicative
        return p * (1 + s) / (1 + p * s)
    def additive_fwd(p, s):
        num = 1 + s + p * s
        dnm = 1 + 2 * p * s
        return p * num / dnm
    def dominant_fwd(p, s):
        num = 1 + s
        dnm = 1 + 2 * p * s * (1 - p) + p * p * s
        return p * num / dnm
    def multiplicative_fwd(p, s):
        num = 1 + s
        dnm = 1 + p * s
        return p * num / dnm
    def recessive_fwd(p, s):
        num = 1 + p * s
        dnm = 1 + p * p * s
        return p * num / dnm

    # one step calculation
    if ploidy == 1:
        one_step = haploid_fwd
    else:
        if one_step_model == 'a':
            one_step = additive_fwd
        elif one_step_model == 'r':
            one_step = recessive_fwd
        elif one_step_model == 'd':
            one_step = dominant_fwd
        else:
            one_step = multiplicative_fwd

    # initialize
    ps = [] # frequencies
    xs = [] # variants
    Ns = [] # sizes
    t = floor(tau0)
    p = pG
    G = max(Ne.keys())
    N = Ne[G]
    x = ceil(p * ploidy * N)
    Ns.append(N)
    xs.append(x)
    ps.append(p)

    if random_walk: # random walk

        while G >= 0:
            G -= 1
            try:
                N = Ne[G]
            except KeyError:
                pass
            if G > t:
                p = one_step(p, s)
            x = int(binom.rvs(int(ploidy * N), p))
            p = x / ploidy / N
            if x < 1:
                break
            if p >= 1:
                break
            ps.append(p)
            xs.append(x)
            Ns.append(N)

        return np.array([np.array(ps), np.array(Ns), np.array(xs)]) # numpy-ify

    else: # deterministic

        while G >= 0:
            G -= 1
            try:
                N = Ne[G]
            except KeyError:
                pass
            if G > t:
                p = one_step(p, s)
            x = floor(p * ploidy * N)
            if x < 1:
                break
            if p >= 1:
                break
            Ns.append(N)
            xs.append(x)
            ps.append(p)

        return np.array([np.array(ps), np.array(Ns), np.array(xs)]) # numpy-ify

### simple coalescent simulators ###

def basic_coalescent(n):
    '''Simulate times in basic coalescent (scale post-hoc by population size)

    Parameters
    ----------
    n : int
        Sample size

    Returns
    -------
    numpy.array
        Interarrival times
    '''
    n = int(float(n))
    k = n
    t = np.zeros(n - 1)
    itr = 0
    curr = 0
    while k > 1:
        bcoef = 2 / k / (k - 1)
        k -= 1
        draw = Exp(bcoef)
        curr += draw
        t[itr] = curr
        itr += 1
    return t

def varying_Ne_coalescent(n, Ne, ploidy = 2, to_tmrca=True):
    '''Simulate times in varying population size coalescent

    Parameters
    ----------
    n : int
        Sample size
    Ne : dict
        Effective population sizes
    ploidy : int
    to_tmrca : bool
        Go to TMRCA

    Returns
    -------
    numpy.array
        Arrival times in generations
    '''

    # initialize
    Ne = zero_shift_Ne(Ne)
    maxG = max(Ne.keys())
    N = Ne[maxG]
    Nt = np.array(list(Ne.values()))
    lambdas = N / Nt # relative (inverse) coalescent rates
    loglambdas = np.log(lambdas)
    Lambdas = np.cumsum(lambdas) # piecewise integral
    Lambdas = np.insert(Lambdas, 0, 0)
    times = np.log(basic_coalescent(n)) + log(ploidy) + log(N) # convert to generations
    times = np.exp(times)
    times = np.insert(times, 0, 0)
    dtime = np.diff(times)
    K = len(dtime)
    Taus = np.zeros(K)
    k = 0
    itr = 0
    LambdaNext = 0
    vCurr = 0
    # while k <= (K-1) means until most recent common ancestor (MRCA)
    # itr < maxG controls updates with piecewise Ne
    while k <= (K-1) and itr < maxG:
        vPrev = vCurr
        LambdaPrev = LambdaCurr = LambdaNext
        RHS = dtime[k] + LambdaPrev
        LambdaNext = RHS
        steps = 0
        try:
            while LambdaCurr < RHS:
                itr += 1
                steps += 1
                LambdaCurr = Lambdas[itr]
            itr -= 1
            boole = int(steps>1)
            LambdaCurr = Lambdas[itr]*boole + LambdaPrev*(1-boole)
            if boole:
                vCurr += exp(log(LambdaCurr-LambdaPrev) - loglambdas[itr-1])
            RHS -= LambdaCurr
            logscalar = loglambdas[itr-1]
            logsoluti = log(RHS) - logscalar
            vCurr += exp(logsoluti)
            Taus[k] = (vCurr - vPrev)
            k += 1
        except IndexError:
            boole = int(steps>1)
            LambdaCurr = Lambdas[-1]*boole + LambdaPrev*(1-boole)
            RHS -= LambdaCurr
            logscalar = loglambdas[-1]
            while steps > 1:
                vCurr += 1
                steps -= 1
            logsoluti = log(RHS) - logscalar
            vCurr += exp(logsoluti)
            Taus[k] = (vCurr - vPrev)
            k += 1
    logscalar = loglambdas[-1]
    if to_tmrca:
        while k <= (K-1): # finish
            Taus[k] = exp(log(dtime[k]) - logscalar)
            k += 1
    Etas = np.cumsum(Taus)
    Etas = Etas[Etas > 0]
#     return np.floor(Etas)+1 # round to discrete time
    return Etas # continuous time

def binomial_step(k, Nei, ploidy=2):
    '''Take a Binomial(n,p) step for 1 gen of WF process

    Parameters
    ----------
    k : int
        Sample size
    Nei : float
        Effective size at time i
    ploidy : int

    Return
    ------
    int
        Count from binomial experiments
    '''
    denom = Nei * ploidy
    numer = int(k * (k - 1) / 2)
    bstep = binom.rvs(numer, 1/denom, size=1)[0]
    return bstep

def binomial_pmf_leq1(k, Nei, ploidy=2):
    '''Compute mass P(\leq 1) for Binomial(n,p) for 1 gen of WF process

    Parameters
    ----------
    k : int
        Sample size
    Nei : float
        Effective size at time i
    ploidy : int

    Return
    ------
    float
        A probability mass
    '''
    numer = int(k * (k - 1) / 2)
    denom = 1 / (Nei * ploidy)
    return binom.pmf(0, numer, denom) + binom.pmf(1, numer, denom)

def wright_fisher(n, Ne, ploidy = 2, to_tmrca=True):
    '''Simulate times in Wright Fisher model with varying population size
    (After last generation in Ne, assume constant size
    and apply scaled basic_coalescent)

    Parameters
    ----------
    n : int
        Sample size
    Ne : dict
        Effective population sizes
    ploidy : int
    to_tmrca : bool
        Continue to TMRCA

    Returns
    -------
    numpy.array
        Arrival times in generations
    '''
    # initalize
    n = int(float(n))
    k = n
    itr = 0
    gentimes = np.zeros(n-1)
    Me = deepcopy(Ne)
    timer = min(Me.keys())
    lastG = max(Me.keys())
    lastN = Me[lastG]
    cuml = 0
    while k > 1 and timer <= lastG:
        # record coalescent events
        size = Me[timer] * ploidy
        draw = [randint(0, size - 1) for i in range(k)]
        table = {}
        for i in range(len(draw)):
            try:
                table[draw[i]] += 1
            except KeyError:
                table[draw[i]] = 1
        events = [val for key, val in table.items() if val >= 2]
        l = sum(events) - len(events)
        cuml += l
        timer += 1
        gentimes[itr:cuml] = timer
        k -= l
        itr += l
    if k > 1: # finish with coalescent
        if to_tmrca:
            finish = basic_coalescent(k) * lastN + lastG
            finish = np.floor(finish) + 1 # round to discrete time
            gentimes[cuml:] = finish
    return gentimes[gentimes > 0]

def simulate_coalgen(n, Ne, ploidy, binom_cut=0.01, to_tmrca=True):
    '''Simulate times in varying size population with large sample.
    Use binomial walk, Wright Fisher process, and the coalescent.

    Parameters
    ----------
    n : int
        Sample size
    Ne : dict
        Effective population sizes
    ploidy : int
    to_tmrca : bool
        Go to TMRCA
    binom_cut : float
        Threshold to conduct binomial update [P(X <= 1) <= \alpha]

    Returns
    -------
    numpy.array
        Arrival times in generations
    '''

    assert n > 1
    assert binom_cut > 0
    assert binom_cut < 1
    n = int(float(n))
    j = n
    Taus = np.zeros(n-1)
    itr = min(Ne.keys()) + 1
    k = 0
    try: # while loop ends, i.e. condition to switch from binomial walk achieved

        # binomial walk process
        while binomial_pmf_leq1(j, Ne[itr], ploidy) <= binom_cut:
            # i is binomial count
            # j is nodes in wf process
            # k is index iterating up
            i = binomial_step(j, Ne[itr], ploidy)
            j -= i
            if j > 1:
                itr += 1
                Taus[k:(k+i)] = itr
                k += i
            else:
                itr += 1
                Taus[k:] = itr
        Taus = Taus[Taus>0]

        # wright fisher / coalescent process
        # wright fisher runs a basic coalescent at end of Ne dictionary
        sNe = {k:v for k,v in Ne.items() if k >= itr}
        Etas = wright_fisher(j, sNe, ploidy, to_tmrca)
        Taus = np.concatenate((Taus,Etas))

    except KeyError: # while loop does not end, i.e. reach end of defined Ne
        Taus = Taus[Taus>0]

    return Taus

### coalescent network simulator ###

import numpy as np
import pandas as pd
import networkx as nx
from scipy.stats import binom
from random import randint, sample
from math import floor, exp, log, ceil
from numpy.random import exponential as Exp

### coalescent network simulator ###

class DirectedPaths:

    def __init__(self, _id, _pairwise_output):
        self._left_length = np.Inf
        self._right_length = np.Inf
        self._num_leaves = 1
        self._leading_leaf = _id
        self._pairwise_output = _pairwise_output
        if _pairwise_output:
            self._leaves = [_id]
        return None

    def add_edge(self, _left_length, _right_length):
        self._left_length = min(self._left_length, _left_length)
        self._right_length = min(self._right_length, _right_length)
        return None

class Node:

    def __init__(self, _id, _time, _pairwise_output):
        self._id = _id
        self._time = _time
        if _time <= 0:
            self._paths_list = [DirectedPaths(_id, _pairwise_output)]
            self._num_leaves = 1
        else:
            self._paths_list = []
            self._num_leaves = 0
        self._num_tracts = 0
        return None
    def __repr__(self):
        return 'Node(_id = ' + str(self._id) + ')'
    def __str__(self):
        return 'Node(_id = ' + str(self._id) + ')'

    def add_incoming_node(self, node):
        self._num_leaves += node._num_leaves
        self._num_tracts += node._num_tracts
        return None

    def add_incoming_edge(self, scale, scales=True):
        if scales: # for multiple compare
            _left_length = Exp(scale=scale, size=1)[0]
            _right_length = Exp(scale=scale, size=1)[0]
            for paths in self._paths_list:
                paths.add_edge(_left_length, _right_length)
        else: # for pair compare
            if scale != 0:
                scale = 1 / scale
                _left_length = Exp(scale=scale, size=1)[0]
                _right_length = Exp(scale=scale, size=1)[0]
                for paths in self._paths_list:
                    paths.add_edge(_left_length, _right_length)
            else:
                pass
        return None

    def merge(self):
        paths_list = self._paths_list
        merged = set() # merge set
        doskip = set() # skip set
        k = len(paths_list)
        for i in range(k):
            if i not in doskip:
                merged.add(i)
                for j in range(i+1,k):
                    if j not in doskip:
                        paths1 = paths_list[i]
                        paths2 = paths_list[j]
                        if paths1._left_length == paths2._left_length:
                            if paths1._right_length == paths2._right_length:
                                m1 = paths1._num_leaves
                                m2 = paths2._num_leaves
                                m = m1 + m2
                                paths_list[i]._num_leaves = m
                                paths_list[i]._leading_leaf = min(paths1._leading_leaf, paths2._leading_leaf)
                                if paths_list[i]._pairwise_output:
                                    paths_list[i]._leaves += paths2._leaves
                                doskip.add(j)
        # merging
        self._paths_list = [paths_list[i] for i in merged]
        return None

    def prune_paths(self, cutoff):
        self._paths_list = [paths for paths in self._paths_list if paths._left_length + paths._right_length >= cutoff]
        return None

    def pair_coalesce(self, node1, node2, cutoff1, cutoff2, subgraph_id=2, record_dist=True, pairwise_output=True):
        deltas = [self._time - node1._time, self._time - node2._time]
        rates = [delta / 100 for delta in deltas]
        node1.add_incoming_edge(rates[0], scales=False)
        node1.prune_paths(cutoff2)
        node1.merge()
        node2.add_incoming_edge(rates[1], scales=False)
        node2.prune_paths(cutoff2)
        node2.merge()
        _num_tracts, n1, n2 = pairwise_compare(node1, node2, cutoff1, cutoff2, self._time, subgraph_id, record_dist, pairwise_output)
        node1 = n1
        node2 = n2
        self._num_tracts += _num_tracts
        self.add_incoming_node(node1)
        for paths in node1._paths_list:
            self._paths_list.append(paths)
        self.add_incoming_node(node2)
        for paths in node2._paths_list:
            self._paths_list.append(paths)
        return None

def pairwise_compare(node1, node2, cutoff1, cutoff2, _time, subgraph_id=2, record_dist=True, pairwise_output=True):
    '''Compare paths up coalescent tree based in IBD

    Parameters
    ----------
    cutoff1 : float
        Long IBD cM threshold
    cutoff2 : float
        Long IBD cM threshold
    _time : int
        Generation time

    Returns
    -------
    tuple
        (int # of long IBD segments, Node class instance, Node class instance)
    '''
    # initializing
    global H1, H0, H, ldist1, ldist0, ldist, tdist1, tdist0, tdist, ddist1, ddist0, ddist
    global pairwise_segments
    _num_tracts = 0
    K = len(node1._paths_list)
    L = len(node2._paths_list)
    for k in range(K):
        i = node1._paths_list[k]
        for l in range(L):
            j = node2._paths_list[l]
            _left_length = min(i._left_length, j._left_length)
            _right_length = min(i._right_length, j._right_length)
            _length = _right_length + _left_length
            # networking
            if _length >= cutoff2:
                if subgraph_id == 1:
                    H1.add_edge(i._leading_leaf, j._leading_leaf)
                elif subgraph_id == 0:
                    H0.add_edge(i._leading_leaf, j._leading_leaf)
                else:
                    H.add_edge(i._leading_leaf, j._leading_leaf)
                if i._leading_leaf <= j._leading_leaf:
                    node2._paths_list[l]._leading_leaf = node1._paths_list[k]._leading_leaf
                else:
                    node1._paths_list[k]._leading_leaf = node2._paths_list[l]._leading_leaf
                # segment counting
                if _length >= cutoff1:
                    m1 = i._num_leaves
                    m2 = j._num_leaves
                    density = m1 * m2
                    _num_tracts += density
                    if record_dist:
                        if subgraph_id == 1:
                            ldist1.append(_length)
                            tdist1.append(_time)
                            ddist1.append(density)
                        elif subgraph_id == 0:
                            ldist0.append(_length)
                            tdist0.append(_time)
                            ddist0.append(density)
                        else:
                            ldist.append(_length)
                            tdist.append(_time)
                            ddist.append(density)
                    if i._pairwise_output:
                        for a in i._leaves:
                            for b in j._leaves:
                                pairwise_segments.append((a,b,_length,_time))

    return _num_tracts, node1, node2

def two_randint(n):
    x = randint(0,n)
    y = randint(0,n)
    while x == y:
        y = randint(0,n)
    return sorted([x,y])

def simulate_ibd(n, Ne, long_ibd=2.0, short_ibd=1.0, ploidy=2, record_dist=True, pairwise_output=True):
    '''ibd segments from a coalescent

    Parameters
    ----------
    n : int
        Sample size (individuals)
    Ne : dict
        Effective population sizes
    long_ibd: float
        cM length threshold
    short_ibd : float
        cM length threshold
    ploidy : int
        1 for haploid or 2 for diploid
    record_dist : bool
        To save tract length and coalescent time distributions or not to (default True)
    pairwise_output : bool
        To save pairwise segments or not to (default True)

    Returns
    -------
    tuple
        (number of tracts, group sizes, length distr., time distr., count distr., pairwise segments)
    '''

    # checks
    assert long_ibd > 0
    assert short_ibd > 0
    assert long_ibd >= short_ibd
    n = int(float(n))
    ploidy = int(float(ploidy))

    # initialize graph network
    global H, H1, H0, ldist, ldist1, ldist0, tdist1, tdist0, tdist, ddist1, ddist0, ddist
    global pairwise_segments
    pairwise_segments = []
    H = nx.Graph()
    ldist = []
    ddist = []
    tdist = []

    # initalize
    m = ploidy*n
    interiors = [Node(i, 0, pairwise_output) for i in range(m)]
    itr = m

    # simulate times
    Ne=cut_Ne(to_max_Ne(fill_Ne(Ne),500),2000)
    times = simulate_coalgen(m, Ne, ploidy, to_tmrca=True)

    indxs = [i for i in range(m)]
    for t in times:
        m -= 1
        new_node = Node(itr, t, pairwise_output)
        sindx = sorted(two_randint(m))
        s1 = sindx[0]
        s2 = sindx[1]
        new_node.pair_coalesce(interiors[s1], interiors[s2], long_ibd, short_ibd, 2, record_dist)
        interiors.append(new_node)
        interiors.pop(s1)
        interiors.pop(s2-1)
        indxs.pop(s1)
        indxs.pop(s2-1)
        indxs.append(itr)
        itr += 1

    numTracts = sum([node._num_tracts for node in interiors])
    ibdGroupSizes = sorted([len(c) for c in nx.connected_components(H)],reverse=True)

    # tract count, group sizes, distributions: lengths, times, density
    return (numTracts,
            np.array(ibdGroupSizes),
            np.array(ldist),
            np.array(tdist),
            np.array(ddist),
            tuple(pairwise_segments)
           )

def simulate_ibd_isweep(n, s, p0, Ne, long_ibd=2.0, short_ibd=1.0, random_walk=True, one_step_model='a', tau0=0, sv=-0.01, ploidy=2, record_dist=True, pairwise_output=True):
    '''ibd segments from a coalescent with selection

    Parameters
    ----------
    n : int
        Sample size (individuals)
    s : float
        Selection coefficient
    p0 : float
        Variant frequency at generation 0
    Ne : dict
        Effective population sizes
    long_ibd, short_ibd : float
        cM length threshold
    random_walk : bool
        True for random walk
    one_step_model : str
        'm', 'a', 'd', or 'r'
    tau0 : int
        Generation when neutrality begins
    sv: float
        Allele frequency of standing variation
        (Default -0.01 will assume de novo sweep)
    ploidy : int
        1 for haploid or 2 for diploid
    record_dist : bool
        To save tract length and coalescent time distributions or not to (default True)
    pairwise_output : bool
        To save pairwise segments or not to (default True)

    Returns
    -------
    tuple(s)
        (all, adaptive allele, non-adaptive allele) then pairwise segments
        Each tuple is (number of tracts, group sizes, length distr., time distr., count distr.)
    '''

    assert ploidy in [1,2]
    assert long_ibd > 0
    assert short_ibd > 0
    assert long_ibd >= short_ibd
    assert one_step_model in ['m','a','r','d']
    assert p0 <= 1
    assert p0 >= 0
    assert sv < 1

    global H, H1, H0, ldist, ldist1, ldist0, tdist1, tdist0, tdist, ddist1, ddist0, ddist
    global pairwise_segments
    pairwise_segments = []
    H1 = nx.Graph()
    H0 = nx.Graph()
    ldist1 = []
    ldist0 = []
    tdist1 = []
    tdist0 = []
    ddist1 = []
    ddist0 = []

    # renaming
    p = p0
    t = tau0
    mdl = one_step_model
    stoc = random_walk
    Ne = cut_Ne(to_max_Ne(fill_Ne(Ne),500),2000)

    # should p0 be fixed
    if (p0 == 0) or (p0 == 1):
        out = simulate_ibd(n, Ne,
                           long_ibd, 
                           short_ibd,
                           ploidy,
                           record_dist, 
                           pairwise_output
                          )
        # numpy-ify
        return (out,
                (np.nan,np.array([]),np.array([]),np.array([]),np.array([])),
                (np.nan,np.array([]),np.array([]),np.array([]),np.array([])),
                tuple(pairwise_segments)
               )

    # calculating structured demographies
    ps, Ns, xs = walk_variant_backward(s, p, Ne, stoc, mdl, tau0, sv, ploidy)
    n = int(float(n))
    n = n * ploidy
    G = max(Ne.keys())
    n1 = floor(n * ps[0])
    n0 = n - n1
    N1 = [ceil(i) for i in Ns * ps]
    N1 = {i:N1[i] for i in range(len(N1))}
    qs = extend_vector(1 - ps, 1, G)
    Ns = np.array(list(Ne.values()))
    N0 = [ceil(i) for i in Ns * qs]
    N0 = {i:N0[i] for i in range(len(N0))}

    # arrival times
    times1 = simulate_coalgen(n1, N1, ploidy, to_tmrca=False)
    interiors1 = [Node(i, 0, pairwise_output) for i in range(n1)]
    interiors0 = [Node(i, 0, pairwise_output) for i in range(n1,n1+n0)]

    # pairwise comparisons pop 1
    m = n1
    indxs = [i for i in range(m)]
    itr = n1 + n0
    for t in times1:
        m -= 1
        new_node = Node(itr, t, pairwise_output)
        sindx = sorted(two_randint(m))
        s1 = sindx[0]
        s2 = sindx[1]
        new_node.pair_coalesce(interiors1[s1], interiors1[s2], long_ibd, short_ibd, 1, record_dist, pairwise_output)
        interiors1.append(new_node)
        interiors1.pop(s1)
        interiors1.pop(s2-1)
        indxs.pop(s1)
        indxs.pop(s2-1)
        indxs.append(itr)
        itr += 1
    numTracts1 = sum([node._num_tracts for node in interiors1])
    ibdGroupSizes1 = sorted([len(c) for c in nx.connected_components(H1)],reverse=True)

    # pairwise comparisons p0
    maxt1=max([node._time for node in interiors1])
    N00 = {key:val for key,val in N0.items() if key <= maxt1}
    N01 = {key:val for key,val in N0.items() if key > maxt1}
    times0 = simulate_coalgen(n0, N00, ploidy, to_tmrca=False)
    m = n0
    indxs = [i for i in range(m)]
    for t in times0:
        m -= 1
        new_node = Node(itr, t, pairwise_output)
        sindx = sorted(two_randint(m))
        s1 = sindx[0]
        s2 = sindx[1]
        new_node.pair_coalesce(interiors0[s1], interiors0[s2], long_ibd, short_ibd, 0, record_dist, pairwise_output)
        interiors0.append(new_node)
        interiors0.pop(s1)
        interiors0.pop(s2-1)
        indxs.pop(s1)
        indxs.pop(s2-1)
        indxs.append(itr)
        itr += 1
    numTracts0 = sum([node._num_tracts for node in interiors0])
    ibdGroupSizes0 = sorted([len(c) for c in nx.connected_components(H0)],reverse=True)

    # arrival times, pairwise comparisons to TMRCA
    H = nx.compose(H1,H0)
    ldist = [*ldist1, *ldist0]
    tdist = [*tdist1, *tdist0]
    ddist = [*ddist1, *ddist0]
    interiors = interiors1 + interiors0
    m = len(interiors)
    indxs = [i for i in range(m)]
    if len(N01.keys()) > 0:
        times2 = simulate_coalgen(m, N01, ploidy, to_tmrca=True)
        for t in times2:
            m -= 1
            new_node = Node(itr, t, pairwise_output)
            sindx = sorted(two_randint(m))
            s1 = sindx[0]
            s2 = sindx[1]
            new_node.pair_coalesce(interiors[s1], interiors[s2], long_ibd, short_ibd, 2, record_dist, pairwise_output)
            interiors.append(new_node)
            interiors.pop(s1)
            interiors.pop(s2-1)
            indxs.pop(s1)
            indxs.pop(s2-1)
            indxs.append(itr)
            itr += 1
    numTracts = sum([node._num_tracts for node in interiors])
    ibdGroupSizes = sorted([len(c) for c in nx.connected_components(H)],reverse=True)

    # numpy-ify
    return ((numTracts,np.array(ibdGroupSizes),np.array(ldist),np.array(tdist),np.array(ddist)),
            (numTracts1,np.array(ibdGroupSizes1),np.array(ldist1),np.array(tdist1),np.array(ddist1)),
            (numTracts0,np.array(ibdGroupSizes0),np.array(ldist0),np.array(tdist0),np.array(ddist0)),
            tuple(pairwise_segments)
           )

### segments ###

# quasi geometric random variables

def probability_quasi_geometric(ps, Ns, ploidy = 2):
    """Probability masses for quasi-geometric coalescent

    Parameters
    ----------
    ps : numpy.array
        Variant frequencies (back in time)
    Ns : numpy.array
        Effective population sizes
    ploidy : int
        1 for haploid or 2 for diploid

    Returns
    -------
    numpy.array
        P(G = g) for generation g; zero element is P(G > g) for some large g
    """

    assert ploidy in [1,2]
    ones = np.array([1 for i in range(len(ps))])
    marg_mass = 1 / (ploidy * ps[:-1] * Ns[:-1])
    marg_mass = np.insert(marg_mass, 0, 0)
    marg_mass = np.minimum(ones, marg_mass)
    surv_mass = ones - marg_mass
    surv_mass = np.cumprod(surv_mass)
    marg_mass = 1 / (ploidy * ps * Ns)
    marg_mass = np.minimum(ones, marg_mass)
    full_mass = surv_mass * marg_mass
    full_mass = np.insert(full_mass, 0, np.maximum(1 - np.sum(full_mass), 0))

    return full_mass

def simulate_quasi_geometric(n, mass):
    """Simulator for quasi-geometric coalescent

    Parameters
    ----------
    n : int
        Simulation size
    mass : numpy.array
        Probability masses

    Returns
    -------
    numpy.array
        Coalescent generation times; np.Inf for P(G > g) for some large g
    """

    n = int(float(n))
    time = np.array([np.Inf] + [i for i in range(1, len(mass))])
    coal = np.random.choice(time, size = n, p = mass)

    return coal

# erlang random variables

def simulate_erlang_segments(geom):
    """Simulator for Erlang (shape=2) ibd segment lengths

    Parameters
    ----------
    geom : numpy.array
        Coalescent generation times (quasi-geometric)

    Returns
    -------
    numpy.array
        Independent ibd segment lengths
    """

    geom = geom[geom != np.Inf]
    m = len(geom)
    left = np.random.exponential(scale = 50 / geom, size = m)
    right = np.random.exponential(scale = 50 / geom, size = m)

    return left + right

def probability_erlang_segments(ab, mass):
    '''Interval probabilities for Erlang (shape=2) ibd lengths

    Parameters
    ----------
    ab : numpy.array
        Increasing floats in centiMorgans
    mass : numpy.array
        P(G = g) for generation g

    Returns
    -------
    numpy.array
        P(\ell \in [u,v)) for many [u,v) intervals
    '''

    # local function
    def puv(u, v, g):

        # error handling
        if u < 0:
            raise ValueError("Left endpoint less than zero")
        if u >= v:
            raise ValueError("Left endpoint exceeds right")

        # integral values
        U = np.exp(- g / 50 * u) * (u * g + 50)
        if np.isinf(v):
            V = 0
        else:
            V = np.exp(- g / 50 * v) * (v * g + 50)

        return (U - V) / 50

    # initialize
    mass = mass[1:]
    geom = np.array([i for i in range(1, len(mass) + 1)])
    M = len(ab)

    # double integral
    puvs = []
    for m in range(1, M):
        pab = puv(ab[m-1], ab[m], geom)
        puvs.append(sum(mass * pab))

    return np.array(puvs)

# probability of ibd

def probability_ibd(ps, Ns, long_ibd = 2, ploidy = 2):
    """Approximate probability of ibd

    Parameters
    ----------
    ps : numpy.array
        Variant frequencies (back in time)
    Ns : numpy.array
        Effective population sizes
    long_ibd : float
        cM length threshold
    ploidy : int
        1 for haploid or 2 for diploid

    Returns
    -------
    float
        approx P(\ell > c) where \ell is ibd length
    """

    assert ploidy in [1,2]

    # initialize
    p0 = ps[0]

    # cumulative products
    with np.errstate(divide = 'ignore'):
        dnm1 = 1 / (ploidy * ps[:-1] * Ns[:-1])
    dnm1 = np.insert(dnm1, 0, 0)
    one = np.array([1 for i in range(len(dnm1))])
    sub1 = np.minimum(one, dnm1)
    onesub1 = one - sub1
    cp1 = np.cumprod(onesub1)

    # products
    with np.errstate(divide = 'ignore'):
        dnm1 = 1 / (ploidy * ps * Ns)
    dnm1 = np.minimum(one, dnm1)

    # partial sum
    pgs = (p0 ** 2) * cp1 * dnm1
    gstar = len(pgs)

    masses = []
    for g in range(1, gstar + 1):
        masses.append((long_ibd * g / 50 + 1) * exp(- g * long_ibd / 50))
    masses = np.array(masses)
    partialp = sum(pgs * masses)

    # approximate coalescent
    if dnm1[-1] >= 0: # check this
        FRp = - np.Inf
    else:
        hp = (p0 ** 2) * cp1[-1] * (1 - dnm1[-1])
        dp = 1 / ploidy / Ns[-1] / ps[-1]
        Cp = log(1 - dp) if not np.isinf(dp) else 0
        CCp = Cp - long_ibd / 50
        FRp = log(-(CCp + long_ibd / 50 * ((gstar + 1) * CCp - 1)))
        FRp = FRp + CCp * (gstar + 1) - log(CCp ** 2)
        FRp = FRp + log(hp) + log(dp) - Cp * (gstar + 1)

    return partialp + exp(FRp)

def probability_ibd_isweep(s, p0, Ne, long_ibd = 2, one_step_model = 'a', tau0 = 0, sv=-0.01, ploidy = 2):
    '''Approximate probability of ibd given a sweep model

    Parameters
    ----------
    s : float
        Selection coefficient
    p0 : float
        Variant frequency at generation 0
    Ne : dict
        Effective population sizes
    long_ibd : float
        cM length threshold
    one_step_model : str
        'm', 'a', 'd', or 'r'
    tau0 : int
        Generation when neutrality begins
    sv: float
        Allele frequency of standing variation
        (Default -0.01 will assume de novo sweep)
    ploidy : int
        1 for haploid or 2 for diploid

    Returns
    -------
    float
        approx P(\ell > c) where \ell is ibd length
    '''
    assert ploidy in [1,2]
    assert p0 >= 0
    assert p0 <= 1
    assert sv < 1
    if p0==1 or p0==0:
        Ne = cut_Ne(to_max_Ne(fill_Ne(Ne),500),1000)
        b = np.array(list(Ne.values()))
        a = extend_vector(np.array([1]),1,max(Ne.keys()))
        d = probability_ibd(a, b, long_ibd, ploidy)
        f = 0
    else:
        Ne = cut_Ne(to_max_Ne(fill_Ne(Ne),500),1000)
        a, b, c = walk_variant_backward(s, p0, Ne, False, one_step_model, tau0, sv, ploidy)
        a = extend_vector(a, a[-1], max(Ne.keys()))
        b = extend_vector(b, b[-1], max(Ne.keys()))
        d = probability_ibd(a, b, long_ibd, ploidy)
        f = probability_ibd(1-a, b, long_ibd, ploidy)
    return d + f

### for asymptotic tests ###

def simulate_ibd_constant(n, Ne, long_ibd=2.0, short_ibd=1.0, ploidy=2, record_dist=True, pairwise_output=True):
    '''ibd segments from a coalescent w/ constant Ne

    Parameters
    ----------
    n : int
        Sample size (individuals)
    Ne : int
        Constant effective population size
    long_ibd, short_ibd : float
        cM length threshold
    ploidy : int
        1 for haploid or 2 for diploid
    record_dist : bool
        To save tract length and coalescent time distributions or not to (default True)
    pairwise_output : bool
        To save pairwise segments or not to (default True)

    Returns
    -------
    tuple
        (number of tracts, group sizes, length distr., time distr., count distr., pairwise segments)
    '''

    # checks
    assert long_ibd > 0
    assert short_ibd > 0
    assert long_ibd >= short_ibd
    n = int(float(n))
    ploidy = int(float(ploidy))

    # initialize graph network
    global H, H1, H0, ldist, ldist1, ldist0, tdist1, tdist0, tdist, ddist1, ddist0, ddist
    global pairwise_segments
    pairwise_segments = []
    H = nx.Graph()
    ldist = []
    ddist = []
    tdist = []

    # initalize
    m = ploidy*n
    interiors = [Node(i, 0, pairwise_output) for i in range(m)]
    itr = m

    # this is the simple change
    # do basic coalescent instead of wright-fisher
    times = basic_coalescent(m) * float(int(Ne)*ploidy)

    indxs = [i for i in range(m)]
    for t in times:
        m -= 1
        new_node = Node(itr, t, pairwise_output)
        sindx = sorted(two_randint(m))
        s1 = sindx[0]
        s2 = sindx[1]
        new_node.pair_coalesce(interiors[s1], interiors[s2], long_ibd, short_ibd, 2, record_dist)
        interiors.append(new_node)
        interiors.pop(s1)
        interiors.pop(s2-1)
        indxs.pop(s1)
        indxs.pop(s2-1)
        indxs.append(itr)
        itr += 1

    numTracts = sum([node._num_tracts for node in interiors])
    ibdGroupSizes = sorted([len(c) for c in nx.connected_components(H)],reverse=True)

    # tract count, group sizes, distributions: lengths, times, density
    return (numTracts,
            np.array(ibdGroupSizes),
            np.array(ldist),
            np.array(tdist),
            np.array(ddist),
            tuple(pairwise_segments)
           )


### time-varying selection ###

def walk_variant_forward_tv(s_s, g_s, pG, Ne, random_walk = False, ploidy = 2):
    '''Variant frequencies forward in time

    Parameters
    ----------
    s_s : list
        List of selection coefficients
    g_s : list
        List of transition times for selection coefficients
        This should be aligned with the s_s parameter
            e.g., s_s=[0.01,0.03,0.02] and g_s=[50,100] means 
            s=0.02 between 100-, s=0.03 between 100-50 and s=0.01 between 50-0
        Should be length one less than s_s
    pG : float
        Variant frequency at maximum generation
    Ne : dict
        Effective population sizes
    random_walk : bool
        True for random walk
    ploidy : int
        1 for haploid or 2 for diploid

    Additive genic selection is assumed

    Returns
    -------
    tuple
        NumPy arrays for frequencies and sizes
    '''

    # local functions
    assert ploidy in [1,2]
    def haploid_fwd(p, s): # haploid is same as multiplicative
        return p * (1 + s) / (1 + p * s)
    def additive_fwd(p, s):
        num = 1 + s + p * s
        dnm = 1 + 2 * p * s
        return p * num / dnm

    # one step calculation
    if ploidy == 1:
        one_step = haploid_fwd
    else:
        one_step = additive_fwd

    # initialize
    ps = [] # frequencies
    xs = [] # variants
    Ns = [] # sizes
    p = pG
    G = max(Ne.keys())
    assert len(s_s) == (len(g_s) + 1)
    s_s2 = deepcopy(s_s)
    g_s2 = deepcopy(g_s)
    s2 = s_s2[-1]
    s_s2 = s_s2[:-1]
    sdict = {g_s2[i]:s_s2[i] for i in range(len(g_s))}
    N = Ne[G]
    x = ceil(p * ploidy * N)
    Ns.append(N)
    xs.append(x)
    ps.append(p)

    if random_walk: # random walk

        while G >= 0:
            G -= 1
            try:
                N = Ne[G]
            except KeyError:
                pass
            try:
                s2 = sdict[G]
            except KeyError:
                pass
            p = one_step(p, s2)
            x = int(binom.rvs(int(ploidy * N), p))
            p = x / ploidy / N
            if x < 1:
                break
            if p >= 1:
                break
            ps.append(p)
            xs.append(x)
            Ns.append(N)

        return np.array([np.array(ps), np.array(Ns), np.array(xs)]) # numpy-ify

    else: # deterministic

        while G >= 0:
            G -= 1
            try:
                N = Ne[G]
            except KeyError:
                pass
            try:
                s2 = sdict[G]
            except KeyError:
                pass
            p = one_step(p, s2)
            x = floor(p * ploidy * N)
            if x < 1:
                break
            if p >= 1:
                break
            Ns.append(N)
            xs.append(x)
            ps.append(p)

        return np.array([np.array(ps), np.array(Ns), np.array(xs)]) # numpy-ify

def walk_variant_backward_tv(s_s, g_s, p0, Ne, random_walk = False, ploidy = 2):
    '''Variant frequencies backward in time

    Parameters
    ----------
    s_s : list
        List of selection coefficients
    g_s : list
        List of transition times for selection coefficients
        This should be aligned with the s_s parameter
            e.g., s_s=[0.01,0.03,0.02] and g_s=[50,100] means 
            s=0.01 between 0-50, s=0.03 between 50-100 and s=0.02 between 100-
        Should be length one less than s_s
    p0 : float
        Variant frequency at generation 0
    Ne : dict
        Effective population sizes
    random_walk : bool
        True for random walk
    ploidy : int
        1 for haploid or 2 for diploid


    Returns
    -------
    tuple
        NumPy arrays for frequencies and sizes
    '''

    # local functions
    assert ploidy in [1,2]
    assert p0 <= 1
    assert p0 >= 0
    def haploid_bwd(p, s): # haploid is same as multiplicative (Felsenstein, 2017)
        return p / (1 + s - s * p)
    def additive_bwd(p, s):
        if s <= 0:
            return p
        a = s
        b = 1 + s - 2 * p * s
        c = - p
        qf = - b + sqrt((b ** 2) - 4 * a * c)
        qf = qf / 2 / a
        return qf

    # one step calculation
    if ploidy == 1:
        one_step = haploid_bwd
    else:
        one_step = additive_bwd

    # initialize
    ps = [] # frequencies
    xs = [] # variants
    Ns = [] # sizes
    p = p0
    N = Ne[0]
    assert len(s_s) == (len(g_s)+1)
    s_s2 = deepcopy(s_s)
    g_s2 = deepcopy(g_s) 
    s2 = deepcopy(s_s2[0])
    s_s2 = s_s2[1:]
    sdict = {g_s2[i]:s_s2[i] for i in range(len(g_s))}
    x = floor(p * ploidy * N)
    Ns.append(N)
    xs.append(x)
    ps.append(p)

    if random_walk: # random walk

        for G in range(1, max(Ne.keys())+1):
            try: # population size change
                N = Ne[G]
            except KeyError:
                pass
            try:
                s2 = sdict[G]
            except KeyError:
                pass
            p = one_step(p, s2)
            x = int(binom.rvs(int(ploidy * N), p))
            p = x / ploidy / N
            if x < 1:
                break
            if p >= 1:
                break
            ps.append(p)
            xs.append(x)
            Ns.append(N)

        return np.array([ps, Ns, xs], dtype=float) # numpy-ify

    else: # deterministic

        for G in range(1, max(Ne.keys())+1):
            try: # population size change
                N = Ne[G]
            except KeyError:
                pass
            try:
                s2 = sdict[G]
            except KeyError:
                pass
            p = one_step(p, s2)
            x = floor(p * ploidy * N)
            if x < 1:
                break
            if p >= 1:
                break
            Ns.append(N)
            xs.append(x)
            ps.append(p)

        return np.array([np.array(ps), np.array(Ns), np.array(xs)]) # numpy-ify
    
def simulate_ibd_isweep_tv(n, 
                           s_s, 
                           g_s, 
                           p0, 
                           Ne, 
                           long_ibd=2.0, 
                           short_ibd=1.0, 
                           random_walk=True, 
                           ploidy=2, 
                           record_dist=True, 
                           pairwise_output=True
                           ):
    '''ibd segments from a coalescent with selection

    Parameters
    ----------
    n : int
        Sample size (individuals)
    s_s : list
        List of selection coefficients
    g_s : list
        List of transition times for selection coefficients
        This should be aligned with the s_s parameter
            e.g., s_s=[0.01,0.03,0.02] and g_s=[50,100] means 
            s=0.01 between 0-50, s=0.03 between 50-100 and s=0.02 between 100-
        Should be length one less than s_s
    p0 : float
        Variant frequency at generation 0
    Ne : dict
        Effective population sizes
    long_ibd, short_ibd : float
        cM length threshold
    random_walk : bool
        True for random walk
    ploidy : int
        1 for haploid or 2 for diploid
    record_dist : bool
        To save tract length and coalescent time distributions or not to (default True)
    pairwise_output : bool
        To save pairwise segments or not to (default True)

    Returns
    -------
    tuple(s)
        (all, adaptive allele, non-adaptive allele) then pairwise segments
        Each tuple is (number of tracts, group sizes, length distr., time distr., count distr.)
    '''

    assert ploidy in [1,2]
    assert long_ibd > 0
    assert short_ibd > 0
    assert long_ibd >= short_ibd
    assert p0 <= 1
    assert p0 >= 0

    global H, H1, H0, ldist, ldist1, ldist0, tdist1, tdist0, tdist, ddist1, ddist0, ddist
    global pairwise_segments
    pairwise_segments = []
    H1 = nx.Graph()
    H0 = nx.Graph()
    ldist1 = []
    ldist0 = []
    tdist1 = []
    tdist0 = []
    ddist1 = []
    ddist0 = []

    # renaming
    p = p0
    stoc = random_walk
    Ne = cut_Ne(to_max_Ne(fill_Ne(Ne),500),2000)

    # should p0 be fixed
    if (p0 == 0) or (p0 == 1):
        out = simulate_ibd(n, Ne,
                           long_ibd, 
                           short_ibd,
                           ploidy,
                           record_dist, 
                           pairwise_output
                          )
        # numpy-ify
        return (out,
                (np.nan,np.array([]),np.array([]),np.array([]),np.array([])),
                (np.nan,np.array([]),np.array([]),np.array([]),np.array([])),
                tuple(pairwise_segments)
               )

    # calculating structured demographies
    ps, Ns, xs = walk_variant_backward_tv(s_s, g_s, p, Ne, stoc, ploidy)
    n = int(float(n))
    n = n * ploidy
    G = max(Ne.keys())
    n1 = floor(n * ps[0])
    n0 = n - n1
    N1 = [ceil(i) for i in Ns * ps]
    N1 = {i:N1[i] for i in range(len(N1))}
    qs = extend_vector(1 - ps, 1, G)
    Ns = np.array(list(Ne.values()))
    N0 = [ceil(i) for i in Ns * qs]
    N0 = {i:N0[i] for i in range(len(N0))}

    # arrival times
    times1 = simulate_coalgen(n1, N1, ploidy, to_tmrca=False)
    interiors1 = [Node(i, 0, pairwise_output) for i in range(n1)]
    interiors0 = [Node(i, 0, pairwise_output) for i in range(n1,n1+n0)]

    # pairwise comparisons pop 1
    m = n1
    indxs = [i for i in range(m)]
    itr = n1 + n0
    for t in times1:
        m -= 1
        new_node = Node(itr, t, pairwise_output)
        sindx = sorted(two_randint(m))
        s1 = sindx[0]
        s2 = sindx[1]
        new_node.pair_coalesce(interiors1[s1], interiors1[s2], long_ibd, short_ibd, 1, record_dist, pairwise_output)
        interiors1.append(new_node)
        interiors1.pop(s1)
        interiors1.pop(s2-1)
        indxs.pop(s1)
        indxs.pop(s2-1)
        indxs.append(itr)
        itr += 1
    numTracts1 = sum([node._num_tracts for node in interiors1])
    ibdGroupSizes1 = sorted([len(c) for c in nx.connected_components(H1)],reverse=True)

    # pairwise comparisons p0
    maxt1=max([node._time for node in interiors1])
    N00 = {key:val for key,val in N0.items() if key <= maxt1}
    N01 = {key:val for key,val in N0.items() if key > maxt1}
    times0 = simulate_coalgen(n0, N00, ploidy, to_tmrca=False)
    m = n0
    indxs = [i for i in range(m)]
    for t in times0:
        m -= 1
        new_node = Node(itr, t, pairwise_output)
        sindx = sorted(two_randint(m))
        s1 = sindx[0]
        s2 = sindx[1]
        new_node.pair_coalesce(interiors0[s1], interiors0[s2], long_ibd, short_ibd, 0, record_dist, pairwise_output)
        interiors0.append(new_node)
        interiors0.pop(s1)
        interiors0.pop(s2-1)
        indxs.pop(s1)
        indxs.pop(s2-1)
        indxs.append(itr)
        itr += 1
    numTracts0 = sum([node._num_tracts for node in interiors0])
    ibdGroupSizes0 = sorted([len(c) for c in nx.connected_components(H0)],reverse=True)

    # arrival times, pairwise comparisons to TMRCA
    H = nx.compose(H1,H0)
    ldist = [*ldist1, *ldist0]
    tdist = [*tdist1, *tdist0]
    ddist = [*ddist1, *ddist0]
    interiors = interiors1 + interiors0
    m = len(interiors)
    indxs = [i for i in range(m)]
    if len(N01.keys()) > 0:
        times2 = simulate_coalgen(m, N01, ploidy, to_tmrca=True)
        for t in times2:
            m -= 1
            new_node = Node(itr, t, pairwise_output)
            sindx = sorted(two_randint(m))
            s1 = sindx[0]
            s2 = sindx[1]
            new_node.pair_coalesce(interiors[s1], interiors[s2], long_ibd, short_ibd, 2, record_dist, pairwise_output)
            interiors.append(new_node)
            interiors.pop(s1)
            interiors.pop(s2-1)
            indxs.pop(s1)
            indxs.pop(s2-1)
            indxs.append(itr)
            itr += 1
    numTracts = sum([node._num_tracts for node in interiors])
    ibdGroupSizes = sorted([len(c) for c in nx.connected_components(H)],reverse=True)

    # numpy-ify
    return ((numTracts,np.array(ibdGroupSizes),np.array(ldist),np.array(tdist),np.array(ddist)),
            (numTracts1,np.array(ibdGroupSizes1),np.array(ldist1),np.array(tdist1),np.array(ddist1)),
            (numTracts0,np.array(ibdGroupSizes0),np.array(ldist0),np.array(tdist0),np.array(ddist0)),
            tuple(pairwise_segments)
           )




# imports
import gzip
import numpy as np
import pandas as pd
from copy import deepcopy

### relabeling, recoding ###

def labeled_allele_frequencies(vcf, labels1):
    '''Process phased diploid VCF data to compute allele frequencies for 2 groups

    Parameters
    ----------
    vcf : str
        Name of phased diploid VCF data
    labels1 : list
        List of haplotype names in group 1

    Return
    -------
    tuple
        (basepair positions, allele frequencies group 1, allele frequencies group 0, ALT allele frequencies)
    '''

    # bring in group 1
    assert type(labels1) is list
    labhap1=labels1

    # compute allele frequencies from phased VCF
    with gzip.open(vcf, 'rt') as f:
        twochar='##'
        while twochar=='##':
            line=f.readline().strip()
            twochar=line[:2]
        header=line
        headerlist=header.split('\t')
        namdip=headerlist[9:]
        namhap1=[nam+'_1' for nam in namdip]
        namhap2=[nam+'_2' for nam in namdip]
        hap11=[i for i in range(len(namhap1)) if namhap1[i] in labhap1]
        hap21=[i for i in range(len(namhap2)) if namhap2[i] in labhap1]
        hap10=[i for i in range(len(namhap1)) if namhap1[i] not in labhap1]
        hap20=[i for i in range(len(namhap2)) if namhap2[i] not in labhap1]
        pos=[]
        count11=[]
        count21=[]
        count10=[]
        count20=[]
        itr=1
        for line in f:
            if (itr % 1000) == 0:
                print(itr)
            splitline=line.split('\t')
            infoline=splitline[:9]
            markline=splitline[9:]
            markhap1=np.array([int(float(mark[0])) for mark in markline])
            markhap2=np.array([int(float(mark[2])) for mark in markline])
            pos.append(int(float(infoline[1])))
            count11.append(markhap1[hap11].sum())
            count21.append(markhap2[hap21].sum())
            count10.append(markhap1[hap10].sum())
            count20.append(markhap2[hap20].sum())
            itr+=1
    denom1=len(hap11)+len(hap21)
    denom0=len(hap10)+len(hap20)
    count1=[(count11[i] + count21[i]) for i in range(len(count11))]
    count0=[(count10[i] + count20[i]) for i in range(len(count11))]
    freq1=np.array(count1)/denom1
    freq0=np.array(count0)/denom0
    denom=denom1+denom0
    freqm=(np.array(count1)+np.array(count0))/denom
    return pos, freq1, freq0, freqm

def putative_allele_frequencies(freq1, freq0, freqm):
    '''Recode allele frequencies to be putative adaptive allele frequencies

    Parameters
    ----------
    freq1 : array-like
    freq0 : array-like
    freqm : array-like
        ALT allele frequencies for group 1
        ALT allele frequencies for group 0
        ALT allele frequencies for both groups

    Return
    ------
    tuple
        array-like putative adaptive allele frequencies
        where putative adaptive is the allele more frequent in group 1 than 0
    '''
    u = []
    v = []
    w = []
    for i in range(len(freq1)):
        y = freq1[i]
        x = freq0[i]
        z = freqm[i]
        if y - z < 0:
            x = 1-x
            y = 1-y
            z = 1-z
        u.append(x)
        v.append(y)
        w.append(z)
    return np.array(v), np.array(u), np.array(w)

def format_allele_table(pos, freq1, freq0, freqm, topn=np.inf):
    '''Create a pandas DataFrame from grouped allele frequencies

    Parameters
    ----------
    pos : array-like
    freq1 : array-like
    freq0 : array-like
    freqm : array-like
        Allele frequencies for group 1, group 0, and both groups
    topn : int
        Number of top positions to study

    Return
    ------
    pandas DataFrame
        Table of adaptive allele frequencies by group
    '''
    v,u,w=putative_allele_frequencies(freq1,freq0,freqm)
    table = {'POS':pos, 'AAF':w, 'AAF1':v, 'AAF0':u, 'DELTA':v-u}
    table = pd.DataFrame(table)
    table['DELTAPRIME'] = table['DELTA'] / np.sqrt(table['AAF'] * (1 - table['AAF']))
    table.sort_values(by=['DELTAPRIME','POS'], inplace=True, ascending=False)
    table.reset_index(inplace=True,drop=True)
    if topn < table.shape[0]:
        return table.loc[:topn,]
    return table

def labeled_allele_frequencies_haploid(vcf, labels1):
    '''Process haploid VCF data to compute allele frequencies for 2 groups

    Parameters
    ----------
    vcf : str
        Name of haploid VCF data
    labels1 : list/str
        Filename for haplotype names in group 1 or a list of them

    Return
    -------
    tuple
        (basepair positions, allele frequencies group 1, allele frequencies group 0, ALT allele frequencies)
    '''

    # bring in group 1
    assert type(labels1) is list
    labhap1=labels1

    # compute allele frequencies from phased VCF
    with gzip.open(vcf, 'rt') as f:
        twochar='##'
        while twochar=='##':
            line=f.readline().strip()
            twochar=line[:2]
        header=line
        headerlist=header.split('\t')
        namdip=headerlist[9:]
        namhap1=[nam+'_1' for nam in namdip]
        hap11=[i for i in range(len(namhap1)) if namhap1[i] in labhap1]
        hap10=[i for i in range(len(namhap1)) if namhap1[i] not in labhap1]
        pos=[]
        count11=[]
        count10=[]
        itr=1
        for line in f:
            if (itr % 1000) == 0:
                print(itr)
            splitline=line.split('\t')
            infoline=splitline[:9]
            markline=splitline[9:]
            markhap1=np.array([int(float(mark[0])) for mark in markline])
            pos.append(int(float(infoline[1])))
            count11.append(markhap1[hap11].sum())
            count10.append(markhap1[hap10].sum())
            itr+=1
    denom1=len(hap11)
    denom0=len(hap10)
    count1=[(count11[i]) for i in range(len(count11))]
    count0=[(count10[i]) for i in range(len(count11))]
    freq1=np.array(count1)/denom1
    freq0=np.array(count0)/denom0
    denom=denom1+denom0
    freqm=(np.array(count1)+np.array(count0))/denom
    return pos, freq1, freq0, freqm



# imports
import numpy as np
import pandas as pd
from copy import deepcopy

### binning ###

def read_bins(file):
    '''Read *.bins file

    Parameters
    ----------
    file : str
        Input file name

    Returns
    -------
    array-like
        Increasing floats in centiMorgans
    '''

    bins = pd.read_csv(file, header = None)
    ab = bins[0]

    return list(ab)

def bin_ibd_segments(ell, ab):
    """Put ibd segments into bins

    Parameters
    ----------
    ell : numpy.array
        ibd segments
    ab : numpy.array
        Increasing floats in centiMorgans

    Returns
    -------
    numpy.array
        Observed counts for ibd segment bins
    """

    M = len(ab)
    obs = []
    for m in range(1, M):
        idx = ell < ab[m]
        obs.append(sum(idx))
        ell = ell[ell >= ab[m]]

    return np.array(obs)

### formatting vectors for plotting ###

def big_format_distribution(distr, counts):
    '''Reformat a vector for plotting with matplotlib, seaborn

    Parameters
    ----------
    distr : array-like
        Vector of realizations (lengths, times, etc.)
    counts : array-like
        Vector of realization multiplicities

    Returns
    -------
    numpy.array
        Adds copies of realization if multiplicity > 1
    '''
    nobs = sum(counts)
    dist = np.zeros(nobs)
    nlen = len(counts)
    itr = 0
    for i in range(nlen):
        ctr = counts[i]
        dis = distr[i]
        while ctr > 0:
            ctr -= 1
            dist[itr] = dis
            itr += 1
    return np.array(dist)

### effective population size ###

from math import floor, exp

def extend_vector(vec, val, mxn):
    '''Extend NumPy array

    Parameters
    ----------
    vec : array-like
        NumPy array
    val : float
        Impute value
    mxn : int
        Array size

    Returns
    -------
    array-like
        NumPy array
    '''
    num = mxn - (len(vec) - 1)
    ext = np.array([val for i in range(num)])
    return np.concatenate((vec, ext))

def zero_shift_Ne(Ne):
    '''Shift the Ne dictionary to start at generation 0

    Parameters
    ----------
    Ne : dict
        Effective population sizes

    Return
    ------
    dict
        Effective population sizes, with keys shifted
    '''
    minG = min(Ne.keys())
    return {(k-minG):v for k,v in Ne.items()}

def cut_Ne(Ne, cut):
    '''Shrink the Ne dictionary to end at specified generational time

    Parameters
    ----------
    Ne : dict
        Effective population sizes
    cut : float
        Generational time

    Returns
    -------
    dict
        Effective population sizes
    '''
    return {k:v for k,v in Ne.items() if k <= cut}

def fill_Ne(Ne):
    '''Fill in effective sizes for non-specified generations

    Parameters
    ----------
    Ne : dict
        Effective population sizes

    Returns
    -------
    dict
        Modified Ne dictionary
    '''
    assert min(Ne.keys())==0
    Me=deepcopy(Ne)
    values=list(Ne.values())
    keys=list(Ne.keys())
    keys=sorted(keys,reverse=True)
    maxkey=max(keys)
    k = 0
    j = 1
    while j <= maxkey:
        prev = values[k]
        try:
            Me[j]
            k += 1
        except KeyError:
            Me[j] = prev
        j += 1
    return dict(sorted(Me.items(),key=lambda x:x[0]))

def to_max_Ne(Ne, maxG = 500):
    '''Fill in effective sizes to a maximum assuming constant in deep past

    Parameters
    ----------
    Ne : dict
        Effective population sizes
    maxG : numeric
        Maximum generation

    Returns
    -------
    Modified Ne dictionary
    '''
    maxG = int(float(maxG))
    lastG = max(Ne.keys())
    Me = deepcopy(Ne)
    if lastG >= maxG:
        pass
    else:
        val = Me[lastG]
        for key in range(lastG+1,maxG+1):
            Me[key] = val
    return Me

def read_Ne(file):
    '''Read *.ne file

    Parameters
    ----------
    file : str
        Input file name

    Returns
    -------
    dict
        dict[generation] = size
    '''

    Ne = {}
    f = open(file, 'r')
    f.readline() # header
    for line in f:
        g, size = line.split('\t')[:2]
        Ne[int(g)] = int(float(size))
    f.close()

    return Ne

def write_Ne(Ne, output_file):
    '''Write Ne dictionary to .ne file

    Parameters
    ----------
    Ne : dict
        Effective population sizes
    output_file : str
        File name

    Returns
    -------
    None
    '''
    with open(output_file, 'w') as f:
        f.write('GEN\tNE\n')
        for k in Ne.keys():
            f.write(str(k)); f.write('\t'); f.write(str(Ne[k])); f.write('\n')
    return None

def extend_Ne(file, maxg, output):
    '''Extend *.ne file

    Constant size until user-specificied generation. Creates a new file.

    Parameters
    ----------
    file : str
        Input file name
    maxg : int
        Maximum generation
    output : str
        Output file name

    Returns
    -------
    None
    '''

    with open(file, 'r') as f:
        lines = f.readlines()
        g, size = lines[-1].strip().split('\t')[:2]

    g = int(g) + 1
    if maxg >= g:
        with open(output, 'w') as o:
            for line in lines:
                o.write(line)
            while maxg >= g:
                o.write(str(g) + '\t' + size + '\n')
                g += 1
    else:
        raise ValueError('*.ne file contains coalescent times greater than maxg')

    return None

def make_constant_Ne(file, size, maxg):
    '''Create *.ne file for constant size population

    Parameters
    ----------
    file: str
        Output file name
    size : float
        Effective population size
    maxg : int
        Maximum generation

    Returns
    -------
    None
    '''

    size = floor(size)
    maxg = floor(maxg)
    with open(file, 'w') as f:
        f.write('GEN\tNE\n')
        f.write(str(0)); f.write('\t'); f.write(str(size)); f.write('\n')
        for i in range(1, int(maxg) + 1):
            f.write(str(i)); f.write('\t'); f.write(str(size)); f.write('\n')

    return None

def make_exponential_Ne(file, size, maxg, rate):
    '''Create *.ne file for exponentially growing population

    Parameters
    ----------
    file : str
        Output file name
    size : float
        Effective population size at generation 0
    maxg : list (int)
        Maximum generation (s)
    rate : list (float)
        Exponential growth rate(s)

    Returns
    -------
    None
    '''

    assert len(rate) == len(maxg)
    size = floor(size)
    with open(file, 'w') as f:
        f.write('GEN\tNE\n')
        f.write(str(0)); f.write('\t'); f.write(str(size)); f.write('\n')
        g1 = 0 + 1
        for j in range(len(rate)):
            g2 = floor(maxg[j])
            r = rate[j]
            for i in range(g1, g2 + 1):
                size = floor(size * exp(r))
                f.write(str(i)); f.write('\t'); f.write(str(size)); f.write('\n')
            g1 = g2 + 1

    return None

def make_bottleneck_Ne(file, isize, bsize, rt1, rt2, rt3):
    '''Create *.ne file for bottlneck population

    Parameters
    ----------
    file : str
        Output file name
    isize : float
        Initial population size
    bsize : float
        Bottleneck population size
    rt1 : tuple
        First phase (exponential rate, time)
    rt2 : tuple
        Second phase (exponential rate, time)
    rt3 : tuple
        Third phase (exponential rate, time)

    Returns
    -------
    None
    '''

    # initialize
    size = []
    r1 = rt1[0]
    t1 = rt1[1]
    r2 = rt2[0]
    t2 = rt2[1]
    r3 = rt3[0]
    t3 = rt3[1]

    # first growth phase
    for i in range(t1+1):
        size.append(isize * exp(r1 * i))

    # second growth phase
    for j in range(1,t2):
        size.append(isize * exp(r1 * t1) * exp(r2 * j))

    # bottleneck
    size.append(bsize)

    # third growth phase
    for k in range(1,t3+1):
        size.append(bsize * exp(r3 * k))

    # output
    size.reverse()
    with open(file, 'w') as f:
        f.write('GEN\tNE\n')
        for l in range(len(size)):
            f.write(str(l))
            f.write('\t')
            f.write(str(size[l]))
            f.write('\n')

    return None

def make_two_growth_phases_Ne(file, isize, rt1, rt2):
    '''Create *.ne file for two growth phases population

    Parameters
    ----------
    file : str
        Output file name
    isize : float
        Initial population size
    rt1 : tuple
        First phase (exponential rate, time)
    rt2 : tuple
        Second phase (exponential rate, time)

    Returns
    -------
    None
    '''

    # initialize
    size = []
    r1 = rt1[0]
    t1 = rt1[1]
    r2 = rt2[0]
    t2 = rt2[1]

    # first growth phase
    for i in range(t1+1):
        size.append(isize * exp(r1 * i))

    # second growth phase
    for j in range(1,t2):
        size.append(isize * exp(r1 * t1) * exp(r2 * j))

    # output
    size.reverse()
    with open(file, 'w') as f:
        f.write('GEN\tNE\n')
        for l in range(len(size)):
            f.write(str(l))
            f.write('\t')
            f.write(str(size[l]))
            f.write('\n')

    return None

def make_three_growth_phases_Ne(file, isize, rt1, rt2, rt3):
    '''Create *.ne file three growth phases population

    Parameters
    ----------
    file : str
        Output file name
    isize : float
        Initial population size
    rt1 : tuple
        First phase (exponential rate, time)
    rt2 : tuple
        Second phase (exponential rate, time)
    rt3 : tuple
        Third phase (exponential rate, time)

    Returns
    -------
    None
    '''

    # initialize
    size = []
    r1 = rt1[0]
    t1 = rt1[1]
    r2 = rt2[0]
    t2 = rt2[1]
    r3 = rt3[0]
    t3 = rt3[1]

    # first growth phase
    for i in range(t1+1):
        size.append(isize * exp(r1 * i))

    # second growth phase
    for j in range(1,t2):
        size.append(isize * exp(r1 * t1) * exp(r2 * j))

    # third growth phase
    for k in range(1,t3+1):
        size.append(isize * exp(r1 * t1) * exp(r2 * j) * exp(r3 * k))

    # output
    size.reverse()
    with open(file, 'w') as f:
        f.write('GEN\tNE\n')
        for l in range(len(size)):
            f.write(str(l))
            f.write('\t')
            f.write(str(size[l]))
            f.write('\n')

    return None



# imports
import gzip
import numpy as np
import pandas as pd
from random import randint
from math import floor, exp, log, ceil
from numpy.random import exponential as Exp
from scipy.stats import norm, binom
from scipy.optimize import minimize, minimize_scalar
from .coalescent import walk_variant_backward, probability_quasi_geometric, probability_erlang_segments
from .utilities import bin_ibd_segments

# read .ibd file

def read_ibd_file(ibd_file, header = 1, include_length = 1):
    '''Create list of IBD segments

    Parameters
    ----------
    ibd_file : str
        Name of text file with IBD segments
    header : int
        0 for no header
    include_length : int
        0 for no length

    Returns
    -------
    list
        (ID1, ID2, cM length) pairs
    '''
    segments = []
    with gzip.open(ibd_file, 'rt') as f:
        if header != 0:
            f.readline()
        for line in f:
            row = line.strip().split('\t')
            id1 = str(row[0]) + '_' + str(row[1])
            id2 = str(row[2]) + '_' + str(row[3])
            if include_length != 0:
                length = float(row[7])
                segments.append((id1, id2, length))
            else:
                segments.append((id1, id2))
    return segments

# goodness-of-fit statistics (estimate selection coeffficent)

def chi2_labeled_isweep(s, p0, Ne, n, obs1, obs0, ab, one_step_model = 'a', tau0 = 0, sv=-0.01, ploidy = 2):
    '''Chi-squared statistic for sweep model (labeled)

    Parameters
    ----------
    s : float
        Selection coefficient
    p0 : float
        Variant frequency at generation 0
    Ne : dict
        Effective population sizes
    n : int
        Pairs sample size
    obs1 : numpy.array
        Observed counts for ibd segment bins (labeled 1)
    obs0 : numpy.array
        Observed counts for ibd segment bins (labeled 0)
    ab : numpy.array
        Increasing floats in centiMorgans
    one_step_model : str
        'm', 'a', 'd', or 'r'
    tau0 : int
        Generation at which neutrality begins
    sv: float
        Allele frequency of standing variation
        (Default -0.01 will assume de novo sweep)
    ploidy : int
        1 for haploid or 2 for diploid

    Returns
    -------
    float
        Goodness-of-fit statistic
    '''

    # assertions
    assert len(obs1) == len(obs0)
    assert len(obs1) == (len(ab) - 1)

    # probabilities
    ps, Ns, xs = walk_variant_backward(s, p0, Ne, False, one_step_model, tau0, sv, ploidy)
    mass1 = (ps[0] ** 2) * probability_quasi_geometric(ps, Ns, ploidy)
    mass0 = ((1 - ps[0]) ** 2) * probability_quasi_geometric(1 - ps, Ns, ploidy)

    # expecteds
    E0 = probability_erlang_segments(ab, mass0)
    E1 = probability_erlang_segments(ab, mass1)
    E1 = n * E1
    E0 = n * E0

    # chi-squared goodness-of-fit
    OE1 = ((obs1 - E1) ** 2) / E1
    OE0 = ((obs0 - E0) ** 2) / E0

    return np.sum(OE1) + np.sum(OE0)

def chi2_isweep(s, p0, Ne, n, obs, ab, one_step_model = 'a', tau0 = 0, sv=-0.01, ploidy = 2):
    '''Chi-squared statistic for sweep model (unlabeled)

    Parameters
    ----------
    s : float
        Selection coefficient
    p0 : float
        Variant frequency at generation 0
    Ne : dict
        Effective population sizes
    n : int
        Pairs sample size
    obs : numpy.array
        Observed counts for ibd segment bins (unlabeled)
    ab : numpy.array
        Increasing floats in centiMorgans
    one_step_model : str
        'm', 'a', 'd', or 'r'
    tau0 : int
        Generation at which neutrality begins
    sv: float
        Allele frequency of standing variation
        (Default -0.01 will assume de novo sweep)
    ploidy : int
        1 for haploid or 2 for diploid

    Returns
    -------
    float
        Goodness-of-fit statistic
    '''

    assert len(obs) == (len(ab) - 1)
    ps, Ns, xs = walk_variant_backward(s, p0, Ne, False, one_step_model, tau0, sv, ploidy)
    mass1 = (ps[0] ** 2) * probability_quasi_geometric(ps, Ns, ploidy)
    mass0 = ((1 - ps[0]) ** 2) * probability_quasi_geometric(1 - ps, Ns, ploidy)
    E0 = probability_erlang_segments(ab, mass0)
    E1 = probability_erlang_segments(ab, mass1)
    E = E1 + E0
    E = n * E
    OE = ((obs - E) ** 2) / E

    return np.sum(OE)

### bootstrap (study uncertainty) ###

def bootstrap_standard(val, boot, alpha1 = 0.025, alpha2 = 0.975):
    '''Implements the standard bootstrap interval estimator

    Parameters
    ----------
    val : float
        Parameter estimate
    boot : numpy.array
        Bootstraps
    alpha1 : float
        Percentile
    alpha2 : float
        Percentile

    Returns
    -------
    tuple
        (lower, middle, upper) interval estimator

    '''
    za = norm.ppf(alpha1)
    zb = norm.ppf(alpha2)
    sd = np.std(boot)

    return (val + za * sd, val, val + zb * sd)

def bootstrap_standard_bc(val, boot, alpha1 = 0.025, alpha2 = 0.975):
    '''Implements the standard bootstrap interval estimator (w/ bias-correction)

    Parameters
    ----------
    val : float
        Parameter estimate
    boot : numpy.array
        Bootstraps
    alpha1 : float
        Percentile
    alpha2 : float
        Percentile

    Returns
    -------
    tuple
        (lower, middle, upper) interval estimator
    '''

    za = norm.ppf(alpha1)
    zb = norm.ppf(alpha2)
    bc = np.mean(boot) - val
    sd = np.std(boot)

    return (val - bc + za * sd, val - bc, val - bc + zb * sd)

def bootstrap_percentile_bc(val, boot, alpha1 = 0.025, alpha2 = 0.975):
    '''Implement percentile-based interval estimator with bias correction

    Parameters
    ----------
    val : float
        Parameter estimate
    boot : numpy.array
        Bootstraps
    alpha1 : float
        Percentile
    alpha2 : float
        Percentile

    Returns
    -------
    tuple
        (lower, middle, upper) interval estimator
    '''
    boot = np.array(boot)
    err = boot - val
    return (
            val - np.quantile(err, alpha2),
            val - np.quantile(err, 0.5),
            val - np.quantile(err, alpha1)
           )

def bootstrap_percentile(val, boot, alpha1 = 0.025, alpha2 = 0.975):
    '''Implements percentile-based interval estimator

    Parameters
    ----------
    val : float
        Parameter estimate
    boot : numpy.array
        Bootstraps
    alpha1 : float
        Percentile
    alpha2 : float
        Percentile

    Returns
    -------
    tuple
        (lower, middle, upper) interval estimator
    '''

    vl = np.quantile(boot, alpha1)
    vu = np.quantile(boot, alpha2)
    vm = np.quantile(boot, 0.5)

    return (vl, vm, vu)

##### time to event #####

def when_count(ct, s, p0, Ne, random_walk = True, one_step_model = 'a', tau0 = 0, sv=-0.01, ploidy=2):
    '''Report when variant count reaches set value

    Parameters
    ----------
    ct : int
        Variant count
    s : float
        Selection coefficient
    p0 : float
        Variant frequency at generation 0
    Ne : dict
        Effective population sizes
    random_walk : bool
        True for random walk
    one_step_model : str
        'm', 'a', 'd', or 'r'
    tau0 : int
        Generation when neutrality begins
    sv: float
        Allele frequency of standing variation
        (Default -0.01 will assume de novo sweep)
    ploidy : int
        1 for haploid or 2 for diploid

    Returns
    -------
    int
        Generation time
    '''

    var = walk_variant_backward(s, p0, Ne, random_walk, one_step_model, tau0, sv, ploidy)[2]
    idx = np.argmax(var <= ct)
    if idx == 0: # handle variant introduction
        return len(var)

    return idx

def bootstrap_count(ct, B, boots, bootp, Ne, random_walk = True, one_step_model = 'a', tau0 = 0, sv=-0.01, ploidy=2):
    '''Parametric bootstrap for variant count generation time

    Parameters
    ----------
    ct : int
        Variant count
    B : int
        Number of bootstraps
    boots : array-like
        NumPy array of bootstraps for selection coefficient
    bootp : array-like
        NumPy array of bootstraps for variant frequency
    Ne : dict
        Effective population sizes
    random_walk : bool
        True for random walk
    one_step_model : str
        'm', 'a', 'd', or 'r'
    tau0 : int
        Generation when neutrality begins
    sv: float
        Allele frequency of standing variation
        (Default -0.01 will assume de novo sweep)
    ploidy : int
        1 for haploid or 2 for diploid

    Returns
    -------
    array-like
        NumPy array of bootstraps for variant count generation time
    '''

    assert len(boots) == len(bootp)
    boott = []
    for i in range(len(boots)):
        for b in range(B):
            boott.append(when_count(ct, boots[i], bootp[i], Ne, random_walk, one_step_model, tau0, sv, ploidy))

    return boott

def when_freq(maf, s, p0, Ne, random_walk = True, one_step_model = 'a', tau0 = 0, sv=-0.01, ploidy=2):
    '''Report when variant frequency reaches set value

    Parameters
    ----------
    maf : float
        Variant frequency
    s : float
        Selection coefficient
    p0 : float
        Variant frequency at generation 0
    Ne : dict
        Effective population sizes
    random_walk : bool
        True for random walk
    one_step_model : str
        'm', 'a', 'd', or 'r'
    tau0 : int
        Generation when neutrality begins
    sv: float
        Allele frequency of standing variation
        (Default -0.01 will assume de novo sweep)
    ploidy : int
        1 for haploid or 2 for diploid

    Returns
    -------
    int
        Generation time
    '''

    var = walk_variant_backward(s, p0, Ne, random_walk, one_step_model, tau0, sv, ploidy)[0]
    idx = np.argmax(var <= maf)
    if idx == 0: # handle variant introduction
        return len(var)

    return idx

def bootstrap_freq(maf, B, boots, bootp, Ne, random_walk = True, one_step_model = 'a', tau0 = 0, sv=-0.01, ploidy=2):
    '''Parametric bootstrap for variant frequency generation time

    Parameters
    ----------
    maf : float
        Variant frequency
    B : int
        Number of bootstraps
    boots : numpy.array
        NumPy array of bootstraps for selection coefficient
    bootp : numpy.array
        NumPy array of bootstraps for variant frequency
    Ne : dict
        Effective population sizes
    random_walk : bool
        True for random walk
    one_step_model : str
        'm', 'a', 'd', or 'r'
    tau0 : int
        Generation when neutrality begins
    sv: float
        Allele frequency of standing variation
        (Default -0.01 will assume de novo sweep)
    ploidy : int
        1 for haploid or 2 for diploid

    Returns
    -------
    numpy.array
        NumPy array of bootstraps for variant frequency generation time
    '''

    assert len(boots) == len(bootp)
    boott = []
    for i in range(len(boots)):
        for b in range(B):
            boott.append(when_freq(maf, boots[i], bootp[i], Ne, random_walk, one_step_model, tau0, sv, ploidy))

    return boott

### metrics (evaluate experimental results) ###

# def check_envelope(x, a, b):
#     '''Check if value is in an interval

#     Parameters
#     ----------
#     x : float
#         Parameter estimate
#     a : float
#         Left endpoint
#     b : float
#         Right endpoint

#     Returns
#     -------
#     bool
#         If covered
#     '''

#     if x >= a:
#         if x <= b:
#             return True

#     return False

# def make_confusion_matrix(real, pred):
#     '''Form confusion matrix for statistical classification

#     Parameters
#     ----------
#     real : array_like
#         Actual 1s and 0s
#     pred : array_like
#         Predicted 1s and 0s

#     Returns
#     -------
#     pandas DataFrame object
#         2 x 2 crosstab
#     '''

#     data = {'Actual':real, 'Predicted':pred}
#     conf = pd.DataFrame(data, columns = ['Actual', 'Predicted'])
#     matr = pd.crosstab(conf['Actual'],
#                        conf['Predicted'],
#                        rownames=['Actual'],
#                        colnames=['Predicted'])

#     return matr

# def false_omission_rate(matrix):
#     '''Compute false omission rate

#     Parameters
#     ----------
#     matrix : pandas DataFrame object
#         2 x 2 cross tab

#     Returns
#     -------
#     float
#         The false omission rate (https://en.wikipedia.org/wiki/Confusion_matrix)
#     '''

#     tn = matrix[0][0]
#     fp = matrix[1][0]
#     fn = matrix[0][1]
#     tp = matrix[1][1]

#     return fn / (fn + tn)

# def false_positive_rate(matrix):
#     '''Compute false positive rate

#     Parameters
#     ----------
#     matrix : pandas DataFrame object
#         2 x 2 cross tab

#     Returns
#     -------
#     float
#         The false positive rate (https://en.wikipedia.org/wiki/Confusion_matrix)
#     '''

#     tn = matrix[0][0]
#     fp = matrix[1][0]
#     fn = matrix[0][1]
#     tp = matrix[1][1]

#     return fp / (fp + tn)

# def true_positive_rate(matrix):
#     '''Compute true positive rate

#     Parameters
#     ----------
#     matrix : pandas DataFrame object
#         2 x 2 cross tab

#     Returns
#     -------
#     float
#         The true positive rate (https://en.wikipedia.org/wiki/Confusion_matrix)
#     '''

#     tn = matrix[0][0]
#     fp = matrix[1][0]
#     fn = matrix[0][1]
#     tp = matrix[1][1]

#     return tp / (tp + fn)



# packages
import sys
import os
import gzip
import numpy as np
import pandas as pd
from scipy.optimize import minimize_scalar, minimize
from scipy.stats import binom, norm
from math import sqrt, floor, ceil, exp, log
from random import randint
from copy import deepcopy
import networkx as nx

# my modules
from .outgroups import *
from .favoredalleles import *
from .utilities import *
from .inference import *
from .coalescent import *
from .slow import *



# Configuration file for the Sphinx documentation builder.

# -- Project information

project = 'isweep'
copyright = '2025, Seth D. Temple'
author = 'Seth D. Temple'

release = '1.0'

# -- General configuration

import os
import sys
sys.path.insert(0, os.path.abspath('../../src'))

extensions = [
    'sphinx.ext.duration',
    'sphinx.ext.doctest',
    'sphinx.ext.autodoc',
    'sphinx.ext.autosummary',
    'sphinx.ext.intersphinx',
    'sphinx.ext.viewcode',
    'sphinx.ext.napoleon',
]

napoleon_google_docstring = False     # Turn on numpydoc strings
napoleon_numpy_docstring = True     # Turn on numpydoc strings

# modules = ['isweep']
autodoc_mock_imports = ['isweep']

intersphinx_mapping = {
    'python': ('https://docs.python.org/3/', None),
    'sphinx': ('https://www.sphinx-doc.org/en/master/', None),
}
intersphinx_disabled_domains = ['std']

templates_path = ['_templates']

# -- Options for HTML output

html_theme = 'sphinx_rtd_theme'

# -- Options for EPUB output
epub_show_urls = 'footnote'

numfig = True
autosectionlabel_prefix_document = True

highlight_language = 'python'



# coalescent ibd simulator for paper
# slow multiple coalesce, wf process based on
# one with on/off buttons for merge, prune
# these are much slower, hence the name slow.py

# importing relevant packages
import numpy as np
import pandas as pd
import networkx as nx
from scipy.stats import binom
from random import randint, sample
from math import floor, exp, log, ceil
from numpy.random import exponential as Exp

from .coalescent import *
from .slow import *

def two_randint(n):
    x = randint(0,n)
    y = randint(0,n)
    while x == y:
        y = randint(0,n)
    return sorted([x,y])

### multiple coalesce, wf process only ###

class WfDirectedPaths:

    def __init__(self, _id):
        self._left_length = np.Inf
        self._right_length = np.Inf
        self._num_leaves = 1
        self._leading_leaf = _id
        return None

    def add_edge(self, _left_length, _right_length):
        self._left_length = min(self._left_length, _left_length)
        self._right_length = min(self._right_length, _right_length)
        return None

class WfNode:

    def __init__(self, _id, _time):
        self._id = _id
        self._time = _time
        if _time <= 0:
            self._paths_list = [WfDirectedPaths(_id)]
            self._num_leaves = 1
        else:
            self._paths_list = []
            self._num_leaves = 0
        self._num_tracts = 0
        return None
    def __repr__(self):
        return 'WfNode(_id = ' + str(self._id) + ')'
    def __str__(self):
        return 'WfNode(_id = ' + str(self._id) + ')'

    def add_incoming_node(self, node):
        self._num_leaves += node._num_leaves
        self._num_tracts += node._num_tracts
        return None

    def add_incoming_edge(self, scale):
        _left_length = Exp(scale=scale, size=1)[0]
        _right_length = Exp(scale=scale, size=1)[0]
        for paths in self._paths_list:
            paths.add_edge(_left_length, _right_length)
        return None

    def merge(self):
        paths_list = self._paths_list
        merged = set() # merge set
        doskip = set() # skip set
        k = len(paths_list)
        for i in range(k):
            if i not in doskip:
                merged.add(i)
                for j in range(i+1,k):
                    if j not in doskip:
                        paths1 = paths_list[i]
                        paths2 = paths_list[j]
                        if paths1._left_length == paths2._left_length:
                            if paths1._right_length == paths2._right_length:
                                m1 = paths1._num_leaves
                                m2 = paths2._num_leaves
                                m = m1 + m2
                                paths_list[i]._num_leaves = m
                                paths_list[i]._leading_leaf = min(paths1._leading_leaf, paths2._leading_leaf)
                                doskip.add(j)
        # merging
        self._paths_list = [paths_list[i] for i in merged]
        return None

    def prune_paths(self, cutoff):
        self._paths_list = [paths for paths in self._paths_list if paths._left_length + paths._right_length >= cutoff]
        return None

    def multiple_coalesce(self, nodes, cutoff1, cutoff2):
        # initializing
        nodes = [node for node in nodes]
        deltas = [self._time - node._time for node in nodes]
        scales = [100 / delta for delta in deltas]
        # updating node info
        for i in range(len(nodes)):
            nodes[i].add_incoming_edge(scales[i])
            nodes[i].prune_paths(cutoff2)
            nodes[i].merge()
        # pairwise comparing
        for i in range(len(nodes)-1):
            for j in range(i+1,len(nodes)):
                _num_tracts, node1, node2 = wf_pairwise_compare(nodes[i], nodes[j], cutoff1, cutoff2)
                nodes[i] = node1
                nodes[j] = node2
                self._num_tracts += _num_tracts
        # passing node info up
        for i in range(len(nodes)):
            self.add_incoming_node(nodes[i])
            for paths in nodes[i]._paths_list:
                self._paths_list.append(paths)
        return None

def wf_pairwise_compare(node1, node2, cutoff1, cutoff2):
    # initializing
    global H
    _num_tracts = 0
    K = len(node1._paths_list)
    L = len(node2._paths_list)
    for k in range(K):
        i = node1._paths_list[k]
        for l in range(L):
            j = node2._paths_list[l]
            _left_length = min(i._left_length, j._left_length)
            _right_length = min(i._right_length, j._right_length)
            _length = _right_length + _left_length
            # networking
            if _length >= cutoff2:
                H.add_edge(i._leading_leaf, j._leading_leaf)
                if i._leading_leaf <= j._leading_leaf:
                    node2._paths_list[l]._leading_leaf = node1._paths_list[k]._leading_leaf
                else:
                    node1._paths_list[k]._leading_leaf = node2._paths_list[l]._leading_leaf
                # segment counting
                if _length >= cutoff1:
                    m1 = i._num_leaves
                    m2 = j._num_leaves
                    density = m1 * m2
                    _num_tracts += density
    return _num_tracts, node1, node2

def simulate_ibd_wf(n, Ne, ibd_min_span=2.0, ibs_min_span=1.0):
    '''ibd segments from a wright fisher process (slower but exact)

    Parameters
    ----------
    n : int
        Haploid sample size
    Ne : dict
        Effective population sizes
    ibd_min_span : float
        cM length threshold
    ibs_min_span : float
        cM length threshold

    Returns
    -------
    list
        Objects of Node class
    '''

    # initialize graph network
    global H
    H = nx.Graph()

    # initalize
    n = int(float(n))
    interiors = [WfNode(i, 0) for i in range(n)]
    itr = n
    timer = 1
    end = max(Ne.keys())
    while timer <= end:
        # record coalescent events
        size = Ne[timer]
        draw = [randint(0, size - 1) for i in range(len(interiors))]
        table = {}
        for i in range(len(draw)):
            try:
                table[draw[i]].append(i)
            except KeyError:
                table[draw[i]] = [i]
        events = [val for key, val in table.items() if len(val) >= 2]
        # record pairwise ibd
        drop = []
        for merge in events:
            # incoporate incoming node data into next level interior node
            itr += 1
            new_node = WfNode(itr, timer)
            nodes = []
            for i in merge:
                nodes.append(interiors[i])
                drop.append(i)
            new_node.multiple_coalesce(nodes, ibd_min_span, ibs_min_span)
            # multiple coalescent, often 2
            interiors.append(new_node)
        # dropping old nodes
        drop = sorted(drop)
        popr = 0
        for d in drop:
            interiors.pop(d + popr)
            popr -= 1
        timer += 1
    numTracts = sum([node._num_tracts for node in interiors])
    maxIbdGroup = max([len(c) for c in nx.connected_components(H)])

    return numTracts, maxIbdGroup

def simulate_ibd_isweep_wf(n, s, p0, Ne, ibd_min_span = 2, ibs_min_span = 1.0, random = True, model = 'm', tau0 = 0, ploidy = 2):
    '''ibd segments from a wright fisher process with selection (slower but exact)

    Parameters
    ----------
    n : int
        Haploid sample size
    s : float
        Selection coefficient
    p0 : float
        Variant frequency at generation 0
    Ne : dict
        Effective population sizes
    ibd_min_span : float
        cM length threshold
    ibs_min_span : float
        cM length threshold
    random : bool
        True for random walk
    model : str
        'm', 'a', 'd', or 'r'
    tau0 : int
        Generation when neutrality begins
    ploidy : int
        Assume diploid

    Returns
    -------
    int
        Count of ibd segments
    '''

    # local function
    def extend_vector(vec, val, mxn):
        '''Extend NumPy array

        Parameters
        ----------
        vec : array-like
            NumPy array
        val : float
            Impute value
        mxn : int
            Array size

        Returns
        -------
        array-like
            NumPy array
        '''
        num = mxn - (len(vec) - 1)
        ext = np.array([val for i in range(num)])
        return np.concatenate((vec, ext))

    global H
    H = nx.Graph()

    # renaming
    p = p0
    t = tau0
    mdl = model
    stoc = random

    # calculating structured demographies
    ps, Ns, xs = walk_variant_backward(s, p, Ne, stoc, mdl, tau0, ploidy)
    n = int(float(n))
    n = n * ploidy
    G = max(Ne.keys())
    n1 = floor(n * ps[0])
    n0 = n - n1
    N1 = [ceil(i) for i in ploidy * Ns * ps]
    N1 = {i:N1[i] for i in range(len(N1))}
    qs = extend_vector(1 - ps, 1, G)
    Ns = np.array(list(Ne.values()))
    N0 = [ceil(i) for i in ploidy * Ns * qs]
    N0 = {i:N0[i] for i in range(len(N0))}

    # population 1
    timer = 1
    end = max(N1.keys())
    itr = n1
    interiors1 = [WfNode(i, 0) for i in range(n1)]
    while timer <= end:
        # record coalescent events
        size = N1[timer]
        draw = [randint(0, size - 1) for i in range(len(interiors1))]
        table = {}
        for i in range(len(draw)):
            try:
                table[draw[i]].append(i)
            except KeyError:
                table[draw[i]] = [i]
        events = [val for key, val in table.items() if len(val) >= 2]
        # record pairwise ibd
        drop = []
        for merge in events:
            # incoporate incoming node data into next level interior node
            itr += 1
            new_node = WfNode(itr, timer)
            nodes = []
            for i in merge:
                nodes.append(interiors1[i])
                drop.append(i)
            new_node.multiple_coalesce(nodes, ibd_min_span, ibs_min_span)
            # multiple coalescent, often 2
            interiors1.append(new_node)
        # dropping old nodes
        drop = sorted(drop)
        popr = 0
        for d in drop:
            interiors1.pop(d + popr)
            popr -= 1
        timer += 1
    numTracts1 = sum([node._num_tracts for node in interiors1])

    # population 0
    timer = 1
    itr += 1
    interiors0 = [WfNode(i, 0) for i in range(itr, itr + n0)]
    itr += n0
    while timer <= end:
        # record coalescent events
        size = N0[timer]
        draw = [randint(0, size - 1) for i in range(len(interiors0))]
        table = {}
        for i in range(len(draw)):
            try:
                table[draw[i]].append(i)
            except KeyError:
                table[draw[i]] = [i]
        events = [val for key, val in table.items() if len(val) >= 2]
        # record pairwise ibd
        drop = []
        for merge in events:
            # incoporate incoming node data into next level interior node
            itr += 1
            new_node = WfNode(itr, timer)
            nodes = []
            for i in merge:
                nodes.append(interiors0[i])
                drop.append(i)
            new_node.multiple_coalesce(nodes, ibd_min_span, ibs_min_span)
            # multiple coalescent, often 2
            interiors0.append(new_node)
        # dropping old nodes
        drop = sorted(drop)
        popr = 0
        for d in drop:
            interiors0.pop(d + popr)
            popr -= 1
        timer += 1
    numTracts0 = sum([node._num_tracts for node in interiors0])

    # pre adaptation
    itr += 1
    end = max(Ne.keys())
    interiors = interiors0 + interiors1
    while timer <= end:
        # record coalescent events
        size = Ne[timer]
        draw = [randint(0, size - 1) for i in range(len(interiors))]
        table = {}
        for i in range(len(draw)):
            try:
                table[draw[i]].append(i)
            except KeyError:
                table[draw[i]] = [i]
        events = [val for key, val in table.items() if len(val) >= 2]
        # record pairwise ibd
        drop = []
        for merge in events:
            # incoporate incoming node data into next level interior node
            itr += 1
            new_node = WfNode(itr, timer)
            nodes = []
            for i in merge:
                nodes.append(interiors[i])
                drop.append(i)
            new_node.multiple_coalesce(nodes, ibd_min_span, ibs_min_span)
            # multiple coalescent, often 2
            interiors.append(new_node)
        # dropping old nodes
        drop = sorted(drop)
        popr = 0
        for d in drop:
            interiors.pop(d + popr)
            popr -= 1
        timer += 1
    numTracts = sum([node._num_tracts for node in interiors])
    maxIbdGroup = max([len(c) for c in nx.connected_components(H)])

    return numTracts, maxIbdGroup, numTracts1, numTracts0

# ### prune, merge on/off buttons ###

# class TimingDirectedPaths:

#     def __init__(self, _id, _pairwise_output):
#         self._left_length = np.Inf
#         self._right_length = np.Inf
#         self._num_leaves = 1
#         self._leading_leaf = _id
#         self._pairwise_output = _pairwise_output
#         if _pairwise_output:
#             self._leaves = [_id]
#         return None

#     def add_edge(self, _left_length, _right_length):
#         self._left_length = min(self._left_length, _left_length)
#         self._right_length = min(self._right_length, _right_length)
#         return None

# class TimingNode:

#     def __init__(self, _id, _time, _pairwise_output):
#         self._id = _id
#         self._time = _time
#         if _time <= 0:
#             self._paths_list = [TimingDirectedPaths(_id, _pairwise_output)]
#             self._num_leaves = 1
#         else:
#             self._paths_list = []
#             self._num_leaves = 0
#         self._num_tracts = 0
#         return None
#     def __repr__(self):
#         return 'TimingNode(_id = ' + str(self._id) + ')'
#     def __str__(self):
#         return 'TimingNode(_id = ' + str(self._id) + ')'

#     def add_incoming_node(self, node):
#         self._num_leaves += node._num_leaves
#         self._num_tracts += node._num_tracts
#         return None

#     def add_incoming_edge(self, scale, scales=True):
#         if scales: # for multiple compare
#             _left_length = Exp(scale=scale, size=1)[0]
#             _right_length = Exp(scale=scale, size=1)[0]
#             for paths in self._paths_list:
#                 paths.add_edge(_left_length, _right_length)
#         else: # for pair compare
#             if scale != 0:
#                 scale = 1 / scale
#                 _left_length = Exp(scale=scale, size=1)[0]
#                 _right_length = Exp(scale=scale, size=1)[0]
#                 for paths in self._paths_list:
#                     paths.add_edge(_left_length, _right_length)
#             else:
#                 pass
#         return None

#     def merge(self):
#         paths_list = self._paths_list
#         merged = set() # merge set
#         doskip = set() # skip set
#         k = len(paths_list)
#         for i in range(k):
#             if i not in doskip:
#                 merged.add(i)
#                 for j in range(i+1,k):
#                     if j not in doskip:
#                         paths1 = paths_list[i]
#                         paths2 = paths_list[j]
#                         if paths1._left_length == paths2._left_length:
#                             if paths1._right_length == paths2._right_length:
#                                 m1 = paths1._num_leaves
#                                 m2 = paths2._num_leaves
#                                 m = m1 + m2
#                                 paths_list[i]._num_leaves = m
#                                 paths_list[i]._leading_leaf = min(paths1._leading_leaf, paths2._leading_leaf)
#                                 if paths_list[i]._pairwise_output:
#                                     paths_list[i]._leaves += paths2._leaves
#                                 doskip.add(j)
#         # merging
#         self._paths_list = [paths_list[i] for i in merged]
#         return None

#     def prune_paths(self, cutoff):
#         self._paths_list = [paths for paths in self._paths_list if paths._left_length + paths._right_length >= cutoff]
#         return None

#     def pair_coalesce(self, node1, node2, cutoff1, cutoff2, subgraph_id=2, record_dist=True, execute_merge=True, execute_prune=True):
#         deltas = [self._time - node1._time, self._time - node2._time]
#         rates = [delta / 100 for delta in deltas]
#         node1.add_incoming_edge(rates[0], scales=False)
#         if execute_prune:
#             node1.prune_paths(cutoff2)
#         if execute_merge:
#             node1.merge()
#         node2.add_incoming_edge(rates[1], scales=False)
#         if execute_prune:
#             node2.prune_paths(cutoff2)
#         if execute_merge:
#             node2.merge()
#         _num_tracts, n1, n2 = timing_pairwise_compare(node1, node2, cutoff1, cutoff2, self._time, subgraph_id, record_dist)
#         node1 = n1
#         node2 = n2
#         self._num_tracts += _num_tracts
#         self.add_incoming_node(node1)
#         for paths in node1._paths_list:
#             self._paths_list.append(paths)
#         self.add_incoming_node(node2)
#         for paths in node2._paths_list:
#             self._paths_list.append(paths)
#         return None

# def timing_pairwise_compare(node1, node2, cutoff1, cutoff2, _time, subgraph_id=2, record_dist=True, pairwise_output=False):
#     # initializing
#     global H1, H0, H, ldist1, ldist0, ldist, tdist1, tdist0, tdist, ddist1, ddist0, ddist
#     global pairwise_segments
#     _num_tracts = 0
#     K = len(node1._paths_list)
#     L = len(node2._paths_list)
#     for k in range(K):
#         i = node1._paths_list[k]
#         for l in range(L):
#             j = node2._paths_list[l]
#             _left_length = min(i._left_length, j._left_length)
#             _right_length = min(i._right_length, j._right_length)
#             _length = _right_length + _left_length
#             # networking
#             if _length >= cutoff2:
#                 if subgraph_id == 1:
#                     H1.add_edge(i._leading_leaf, j._leading_leaf)
#                 elif subgraph_id == 0:
#                     H0.add_edge(i._leading_leaf, j._leading_leaf)
#                 else:
#                     H.add_edge(i._leading_leaf, j._leading_leaf)
#                 if i._leading_leaf <= j._leading_leaf:
#                     node2._paths_list[l]._leading_leaf = node1._paths_list[k]._leading_leaf
#                 else:
#                     node1._paths_list[k]._leading_leaf = node2._paths_list[l]._leading_leaf
#                 # segment counting
#                 if _length >= cutoff1:
#                     m1 = i._num_leaves
#                     m2 = j._num_leaves
#                     density = m1 * m2
#                     _num_tracts += density
#                     if record_dist:
#                         if subgraph_id == 1:
#                             ldist1.append(_length)
#                             tdist1.append(_time)
#                             ddist1.append(density)
#                         elif subgraph_id == 0:
#                             ldist0.append(_length)
#                             tdist0.append(_time)
#                             ddist0.append(density)
#                         else:
#                             ldist.append(_length)
#                             tdist.append(_time)
#                             ddist.append(density)
#                     if i._pairwise_output:
#                         for a in i._leaves:
#                             for b in j._leaves:
#                                 pairwise_segments.append((a,b,_length,_time))

#     return _num_tracts, node1, node2

# def simulate_ibd_timing(n, Ne, long_ibd=2.0, short_ibd=1.0, ploidy=2, record_dist=False, pairwise_output=False, execute_merge=True, execute_prune=True):
#     '''ibd segments from a coalescent

#     Parameters
#     ----------
#     n : int
#         Sample size (individuals)
#     Ne : dict
#         Effective population sizes
#     long_ibd, short_ibd : float
#         cM length threshold
#     ploidy : int
#         1 for haploid or 2 for diploid
#     continuous_time : bool
#         To implement coalescent or Wright-Fisher (default True is coalescent)
#     record_dist : bool
#         To save tract length and coalescent time distributions or not to (default False)
#     pairwise_output : bool
#         To save pairwise segments or not to (default False)
#     execute_merge : bool
#         To implement algorithm with merging or not to (default True)
#     execute_prune : bool
#         To implement algorithm with pruning or not to (default True)

#     Returns
#     -------
#     tuple
#         (number of tracts, group sizes, length distr., time distr., count distr., pairwise segments)
#     '''

#     # checks
#     assert long_ibd > 0
#     assert short_ibd > 0
#     assert long_ibd >= short_ibd
#     n = int(float(n))
#     ploidy = int(float(ploidy))

#     # initialize graph network
#     global H, ldist, tdist, ddist
#     global pairwise_segments
#     pairwise_segments = []
#     H = nx.Graph()
#     ldist = []
#     ddist = []
#     tdist = []

#     # initalize
#     m = ploidy*n
#     interiors = [TimingNode(i, 0, pairwise_output) for i in range(m)]
#     itr = m

#     # simulate times
#     Ne=cut_Ne(to_max_Ne(fill_Ne(Ne),500),2000)
#     times = varying_Ne_coalescent(m, Ne, ploidy, to_tmrca=True)

#     indxs = [i for i in range(m)]
#     for t in times:
#         m -= 1
#         new_node = TimingNode(itr, t, pairwise_output)
#         sindx = sorted(two_randint(m))
#         s1 = sindx[0]
#         s2 = sindx[1]
#         new_node.pair_coalesce(interiors[s1], interiors[s2], long_ibd, short_ibd, 2, record_dist, execute_merge, execute_prune)
#         interiors.append(new_node)
#         interiors.pop(s1)
#         interiors.pop(s2-1)
#         indxs.pop(s1)
#         indxs.pop(s2-1)
#         indxs.append(itr)
#         itr += 1

#     numTracts = sum([node._num_tracts for node in interiors])
#     ibdGroupSizes = sorted([len(c) for c in nx.connected_components(H)],reverse=True)

#     # tract count, group sizes, distributions: lengths, times, density
#     return (numTracts,
#             np.array(ibdGroupSizes),
#             np.array(ldist),
#             np.array(tdist),
#             np.array(ddist),
#             tuple(pairwise_segments)
#            )

# def simulate_ibd_isweep_timing(n, s, p0, Ne, long_ibd=2.0, short_ibd=1.0, random_walk=True, one_step_model='a', tau0=0, ploidy=2, record_dist=False, pairwise_output=False, execute_merge=True, execute_prune=True):
#     '''ibd segments from a coalescent with selection

#     Parameters
#     ----------
#     n : int
#         Sample size (individuals)
#     s : float
#         Selection coefficient
#     p0 : float
#         Variant frequency at generation 0
#     Ne : dict
#         Effective population sizes
#     long_ibd, short_ibd : float
#         cM length threshold
#     random_walk : bool
#         True for random walk
#     one_step_model : str
#         'm', 'a', 'd', or 'r'
#     tau0 : int
#         Generation when neutrality begins
#     ploidy : int
#         1 for haploid or 2 for diploid
#     continuous_time : bool
#         To implement coalescent or Wright-Fisher (default True is coalescent)
#     record_dist : bool
#         To save tract length and coalescent time distributions or not to (default False)
#     pairwise_output : bool
#         To save pairwise segments or not to (default False)
#     execute_merge : bool
#         To implement algorithm with merging or not to (default True)
#     execute_prune : bool
#         To implement algorithm with pruning or not to (default True)

#     Returns
#     -------
#     tuple(s)
#         (all, adaptive allele, non-adaptive allele) then pairwise segments
#         Each tuple is (number of tracts, group sizes, length distr., time distr., count distr.)
#     '''

#     assert ploidy in [1,2]
#     assert long_ibd > 0
#     assert short_ibd > 0
#     assert long_ibd >= short_ibd
#     assert one_step_model in ['m','a','r','d']
#     assert p0 <= 1
#     assert p0 >= 0

#     global H, H1, H0, ldist, ldist1, ldist0, tdist1, tdist0, tdist, ddist1, ddist0, ddist
#     global pairwise_segments
#     pairwise_segments = []
#     H1 = nx.Graph()
#     H0 = nx.Graph()
#     ldist1 = []
#     ldist0 = []
#     tdist1 = []
#     tdist0 = []
#     ddist1 = []
#     ddist0 = []

#     # renaming
#     p = p0
#     t = tau0
#     mdl = one_step_model
#     stoc = random_walk
#     Ne = cut_Ne(to_max_Ne(fill_Ne(Ne),500),1000)

#     # should p0 be fixed
#     if (p0 == 0) or (p0 == 1):
#         out = simulate_ibd_timing(n, Ne,
#                            long_ibd, short_ibd,
#                            ploidy, continuous_time,
#                            record_dist, pairwise_output,
#                            execute_merge, execute_prune
#                           )
#         # numpy-ify
#         return (out,
#                 (np.nan,np.array([]),np.array([]),np.array([]),np.array([])),
#                 (np.nan,np.array([]),np.array([]),np.array([]),np.array([])),
#                 tuple(pairwise_segments)
#                )

#     # calculating structured demographies
#     ps, Ns, xs = walk_variant_backward(s, p, Ne, stoc, mdl, tau0, ploidy)
#     n = int(float(n))
#     n = n * ploidy
#     G = max(Ne.keys())
#     n1 = floor(n * ps[0])
#     n0 = n - n1
#     N1 = [ceil(i) for i in Ns * ps]
#     N1 = {i:N1[i] for i in range(len(N1))}
#     qs = extend_vector(1 - ps, 1, G)
#     Ns = np.array(list(Ne.values()))
#     N0 = [ceil(i) for i in Ns * qs]
#     N0 = {i:N0[i] for i in range(len(N0))}

#     # arrival times
#     times1 = varying_Ne_coalescent(n1, N1, ploidy, to_tmrca=False)

#     interiors1 = [TimingNode(i, 0, pairwise_output) for i in range(n1)]
#     interiors0 = [TimingNode(i, 0, pairwise_output) for i in range(n1,n1+n0)]

#     # pairwise comparisons pop 1
#     m = n1
#     indxs = [i for i in range(m)]
#     itr = n1 + n0
#     for t in times1:
#         m -= 1
#         new_node = TimingNode(itr, t, pairwise_output)
#         sindx = sorted(two_randint(m))
#         s1 = sindx[0]
#         s2 = sindx[1]
#         new_node.pair_coalesce(interiors1[s1], interiors1[s2], long_ibd, short_ibd, 1, record_dist, execute_merge, execute_prune)
#         interiors1.append(new_node)
#         interiors1.pop(s1)
#         interiors1.pop(s2-1)
#         indxs.pop(s1)
#         indxs.pop(s2-1)
#         indxs.append(itr)
#         itr += 1
#     numTracts1 = sum([node._num_tracts for node in interiors1])
#     ibdGroupSizes1 = sorted([len(c) for c in nx.connected_components(H1)],reverse=True)

#     # pairwise comparisons p0
#     maxt1=max([node._time for node in interiors1])
#     N01 = {key:val for key,val in N0.items() if key <= maxt1}
#     N00 = {key:val for key,val in N0.items() if key > maxt1}
#     times0 = varying_Ne_coalescent(n0, N01, ploidy, to_tmrca=False)
#     m = n0
#     indxs = [i for i in range(m)]
#     for t in times0:
#         m -= 1
#         new_node = TimingNode(itr, t, pairwise_output)
#         sindx = sorted(two_randint(m))
#         s1 = sindx[0]
#         s2 = sindx[1]
#         new_node.pair_coalesce(interiors0[s1], interiors0[s2], long_ibd, short_ibd, 0, record_dist, execute_merge, execute_prune)
#         interiors0.append(new_node)
#         interiors0.pop(s1)
#         interiors0.pop(s2-1)
#         indxs.pop(s1)
#         indxs.pop(s2-1)
#         indxs.append(itr)
#         itr += 1
#     numTracts0 = sum([node._num_tracts for node in interiors0])
#     ibdGroupSizes0 = sorted([len(c) for c in nx.connected_components(H0)],reverse=True)

#     # arrival times, pairwise comparisons to TMRCA
#     H = nx.compose(H1,H0)
#     ldist = [*ldist1, *ldist0]
#     tdist = [*tdist1, *tdist0]
#     ddist = [*ddist1, *ddist0]
#     interiors = interiors1 + interiors0
#     m = len(interiors)
#     indxs = [i for i in range(m)]
#     if len(N00.keys()) > 0:
#         times2 = varying_Ne_coalescent(m, N00, ploidy, to_tmrca=True)
#         for t in times2:
#             m -= 1
#             new_node = TimingNode(itr, t, pairwise_output)
#             sindx = sorted(two_randint(m))
#             s1 = sindx[0]
#             s2 = sindx[1]
#             new_node.pair_coalesce(interiors[s1], interiors[s2], long_ibd, short_ibd, 2, record_dist, execute_merge, execute_prune)
#             interiors.append(new_node)
#             interiors.pop(s1)
#             interiors.pop(s2-1)
#             indxs.pop(s1)
#             indxs.pop(s2-1)
#             indxs.append(itr)
#             itr += 1
#     numTracts = sum([node._num_tracts for node in interiors])
#     ibdGroupSizes = sorted([len(c) for c in nx.connected_components(H)],reverse=True)

#     # numpy-ify
#     return ((numTracts,np.array(ibdGroupSizes),np.array(ldist),np.array(tdist),np.array(ddist)),
#             (numTracts1,np.array(ibdGroupSizes1),np.array(ldist1),np.array(tdist1),np.array(ddist1)),
#             (numTracts0,np.array(ibdGroupSizes0),np.array(ldist0),np.array(tdist0),np.array(ddist0)),
#             tuple(pairwise_segments)
#            )



# imports
import numpy as np
import networkx as nx
from copy import deepcopy

def make_ibd_graph(ibd_segments):
    '''Create IBD graph from pairwise segments

    Parameters
    ----------
    ibd_segments : tuple
        Collection of pairwise segments

    Returns
    -------
    networkx.Graph
        Graph w/ edges if haplotypes have a detectable IBD segment
    '''
    ibdgraph=nx.Graph()
    edgelist=[]
    for edge in ibd_segments:
        edgelist.append((edge[0],edge[1]))
    ibdgraph.add_edges_from(edgelist)
    return ibdgraph

def diameter_communities(graph, K=3, max_communities=np.inf):
    '''Method to find connected communities with max diameter 2*K

    Parameters
    ----------
    graph : networkx.Graph
    K : int
        Default is max diameter 2*3
    max_communities : int
        Default is to find all communities

    Returns
    -------
    list
        Node sets for fully connected subgraphs with max diameter 2*K
    '''

    # initialize community detection loop
    copygraph = deepcopy(graph)
    communities = []
    stayput=True
    if max_communities >= np.inf:
        max_communities = copygraph.number_of_nodes()

    # initialize degree centrality
    M = copygraph.number_of_nodes() - 1
    degrees = nx.degree_centrality(copygraph)
    degrees = {key:round(val*M,0) for key, val in degrees.items()}
    sorted_degrees = sorted([(key, val) for key, val in degrees.items()], key=lambda x:x[1], reverse=True)

    for i in range(max_communities):

        # find community center
        community = dict()
        while stayput:
            try:
                center = sorted_degrees[0][0]
                if copygraph.has_node(center):
                    stayput=False
                else:
                    sorted_degrees.pop(0)
            except:
                return communities
        current_neighbors = [center]
        community[center] = 0

        # find up to K-order neighbors
        k = 1
        while k <= K: # up to K-order neighbors
            new_neighbors = []
            for neighbor in current_neighbors:
                adjacent_neighbors = copygraph.neighbors(neighbor)
                for adjacent_neighbor in adjacent_neighbors:
                    if adjacent_neighbor not in community:
                        community[adjacent_neighbor] = 0
                        new_neighbors.append(adjacent_neighbor)
                copygraph.remove_node(neighbor)
            current_neighbors = new_neighbors
            k += 1
        for neighbor in current_neighbors:
            if copygraph.has_node(neighbor):
                copygraph.remove_node(neighbor)
        communities.append(set(community.keys()))
        stayput=True

    return communities

def community_list(graph, communities):
    '''Return list of subgraphs

    Parameters
    ----------
    graph : networkx.Graph
    communities : list of sets
    '''
    return [graph.subgraph(comm) for comm in communities]

def largest_community(communities):
    ''' Return haploids in largest community

    Parameters
    ----------
    list
        Haplotype/sample IDs in the largest community
    '''
    return list(communities[0])

def outlier_communities(communities, scalar=5):
    '''Compute sum of # nodes in outlier big communities

    Parameters
    ----------
    communities : list
    scalar : float
        Multiplier of numpy.std()

    Returns
    -------
    list
        IDs for haploids in outlier communities
    '''
    community_sizes = [len(community) for community in communities]
    community_sizes = np.array(community_sizes)
    cutoff = community_sizes.mean() + community_sizes.std() * scalar
    oindices = (community_sizes > cutoff).nonzero()
    outliers = set()
    for i in oindices[0]:
        community = communities[i]
        outliers = community.union(outliers)
    return list(outliers)
